#!/usr/bin/env python3

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

############################################################
# Density estimation
#
# Compute a density estimate for a log-normal process measured by a
# Poissonian likelihood.
#
# Demo takes a while to compute
#############################################################

import nifty.cl as ift

if __name__ == "__main__":
    filename = "getting_started_density_{}.png"
    ift.random.push_sseq_from_seed(42)

    # Set up signal domain
    npix1 = 128
    npix2 = 128
    sp1 = ift.RGSpace(npix1)
    sp2 = ift.RGSpace(npix2)
    position_space = ift.DomainTuple.make((sp1, sp2))

    signal, ops = ift.density_estimator(position_space)
    correlated_field = ops["correlated_field"]

    data_space = signal.target

    # Generate mock signal and data
    rng = ift.random.current_rng()
    mock_position = ift.from_random(signal.domain, "normal")
    data = ift.Field.from_raw(data_space, rng.poisson(signal(mock_position).asnumpy()))

    # Rejoin domains for plotting
    plotting_domain = ift.DomainTuple.make(ift.RGSpace((npix1, npix2)))
    plotting_domain_expanded = ift.DomainTuple.make(ift.RGSpace((2 * npix1, 2 * npix2)))

    plot = ift.Plot()
    plot.add(
        ift.Field.from_raw(
            plotting_domain_expanded, ift.exp(correlated_field(mock_position)).asnumpy()
        ),
        title="Pre-Slicing Truth",
    )
    plot.add(
        ift.Field.from_raw(plotting_domain, signal(mock_position).asnumpy()),
        title="Ground Truth",
    )
    plot.add(ift.Field.from_raw(plotting_domain, data.asnumpy()), title="Data")
    plot.output(ny=1, nx=3, xsize=10, ysize=3, name=filename.format("setup"))
    print("Setup saved as", filename.format("setup"))

    # Minimization parameters
    ic_sampling = ift.AbsDeltaEnergyController(
        name="Sampling", deltaE=0.01, iteration_limit=100
    )
    ic_newton = ift.AbsDeltaEnergyController(
        name="Newton", deltaE=0.01, iteration_limit=35
    )
    ic_sampling.enable_logging()
    ic_newton.enable_logging()
    minimizer = ift.NewtonCG(ic_newton, enable_logging=True)

    # Number of samples used to estimate the KL
    n_samples = 5

    # Set up likelihood energy and information Hamiltonian
    likelihood_energy = ift.PoissonianEnergy(data) @ signal
    ham = ift.StandardHamiltonian(likelihood_energy, ic_sampling)

    # Start minimization
    initial_mean = 1e-2 * ift.from_random(ham.domain)
    mean = initial_mean

    for i in range(3):
        # Draw new samples and minimize KL
        kl = ift.SampledKLEnergy(mean, ham, n_samples, None)
        kl, convergence = minimizer(kl)
        mean = kl.position

        # Plot current reconstruction
        plot = ift.Plot()
        plot.add(
            ift.Field.from_raw(
                plotting_domain_expanded, ift.exp(correlated_field(mock_position)).asnumpy()
            ),
            title="Ground truth",
        )
        plot.add(
            ift.Field.from_raw(plotting_domain, signal(mock_position).asnumpy()),
            title="Ground truth",
        )
        plot.add(
            ift.Field.from_raw(plotting_domain, kl.samples.average(signal).asnumpy()),
            title="Reconstruction",
        )
        plot.add(
            (ic_newton.history, ic_sampling.history, minimizer.inversion_history),
            label=["kl", "Sampling", "Newton inversion"],
            title="Cumulative energies",
            s=[None, None, 1],
            alpha=[None, 0.2, None],
        )
        plot.output(
            nx=3, ny=2, ysize=10, xsize=15, name=filename.format(f"loop_{i:02d}")
        )

    # Done, compute posterior statistics
    mean, var = kl.samples.sample_stat(signal)
    mean_unsliced, var_unsliced = kl.samples.sample_stat(correlated_field.exp())

    # Plotting
    plot = ift.Plot()
    plot.add(ift.Field.from_raw(plotting_domain, mean.asnumpy()), title="Posterior Mean")
    plot.add(
        ift.Field.from_raw(plotting_domain, ift.sqrt(var).asnumpy()),
        title="Posterior Standard Deviation",
    )
    plot.add(
        ift.Field.from_raw(plotting_domain_expanded, mean_unsliced.asnumpy()),
        title="Posterior Unsliced Mean",
    )
    plot.add(
        ift.Field.from_raw(plotting_domain_expanded, ift.sqrt(var_unsliced).asnumpy()),
        title="Posterior Unsliced Standard Deviation",
    )
    plot.output(xsize=15, ysize=15, name=filename.format("results"))
    print("Saved results as", filename.format("results"))



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial
from math import log2, sqrt
from typing import Callable, Iterable, Literal, Optional, Tuple, Union

from jax import numpy as jnp
import numpy as np

from .util import (
    coarse2fine_distances,
    coarse2fine_shape,
    fine2coarse_distances,
    fine2coarse_shape,
)

DEPTH_RANGE = (0, 32)
MAX_SIZE0 = 1024

NEST = True


class CoordinateChart:
    def __init__(
        self,
        min_shape: Optional[Iterable[int]] = None,
        depth: Optional[int] = None,
        *,
        shape0: Optional[Iterable[int]] = None,
        _coarse_size: int = 5,
        _fine_size: int = 4,
        _fine_strategy: Literal["jump", "extend"] = "extend",
        rg2cart: Optional[
            Callable[
                [
                    Iterable,
                ],
                Iterable,
            ]
        ] = None,
        cart2rg: Optional[
            Callable[
                [
                    Iterable,
                ],
                Iterable,
            ]
        ] = None,
        regular_axes: Optional[Union[Iterable[int], Tuple]] = None,
        irregular_axes: Optional[Union[Iterable[int], Tuple]] = None,
        distances: Optional[Union[Iterable[float], float]] = None,
        distances0: Optional[Union[Iterable[float], float]] = None,
    ):
        """Initialize a refinement chart.

        Parameters
        ----------
        min_shape :
            Minimal extent in pixels along each axes at the final refinement
            level.
        depth :
            Number of refinement iterations.
        shape0 :
            Alternative to `min_shape` and specifies the extent in pixels along
            each axes at the zeroth refinement level.
        _coarse_size :
            Number of coarse pixels which to refine to `_fine_size` fine
            pixels.
        _fine_size :
            Number of fine pixels which to refine from `_coarse_size` coarse
            pixels.
        _fine_strategy :
            Whether to space fine pixels solely within the centermost coarse
            pixel ("jump"), or whether to always space them out s.t. each fine
            pixels takes up half the Euclidean volume of a coarse pixel
            ("extend").
        rg2cart :
            Function to translate Euclidean points on a regular coordinate
            system to the Cartesian coordinate system of the modeled points.
        cart2rg :
            Inverse of `rg2cart`.
        regular_axes :
            Informs the coordinate chart on symmetries within the Cartesian
            coordinate system of the modeled points. If specified, refinement
            matrices are broadcasted as need instead of recomputed.
        irregular_axes :
            Negative of `regular_axes`. Specifying either is sufficient.
        distances :
            Special case of a coordinate chart in which the regular grid points
            are merely stretched or compressed. `distances` are used to set the
            distance between points along every axes at the final refinement
            level.
        distances0:
            Same as `distances` except that `distances0` refers to the
            distances along every axes at the zeroth refinement level.

        Note
        ----
        The functions `rg2cart` and `cart2rg` are always w.r.t. the grid at
        zero depth. In other words, it is straight forward to increase the
        resolution of an existing chart by simply increasing its depth.
        However, extending a grid spatially is more cumbersome and is best done
        via `shape0`.
        """
        if depth is None:
            if min_shape is None:
                raise ValueError("specify `min_shape` to infer `depth`")
            if shape0 is not None or distances0 is not None:
                ve = "can not infer `depth` with `shape0` or `distances0` set"
                raise ValueError(ve)
            for depth in range(*DEPTH_RANGE):
                shape0 = fine2coarse_shape(
                    min_shape,
                    depth=depth,
                    ceil_sizes=True,
                    _coarse_size=_coarse_size,
                    _fine_size=_fine_size,
                    _fine_strategy=_fine_strategy,
                )
                if np.prod(shape0, dtype=int) <= MAX_SIZE0:
                    break
            else:
                ve = f"unable to find suitable `depth`; please specify manually"
                raise ValueError(ve)
        if depth < 0:
            raise ValueError(f"invalid `depth`; got {depth!r}")
        self._depth = depth

        if shape0 is None and min_shape is not None:
            shape0 = fine2coarse_shape(
                min_shape,
                depth,
                ceil_sizes=True,
                _coarse_size=_coarse_size,
                _fine_size=_fine_size,
                _fine_strategy=_fine_strategy,
            )
        elif shape0 is None:
            raise ValueError("either `shape0` or `min_shape` must be specified")
        self._shape0 = (shape0,) if isinstance(shape0, int) else tuple(shape0)
        self._shape = coarse2fine_shape(
            shape0,
            depth,
            _coarse_size=_coarse_size,
            _fine_size=_fine_size,
            _fine_strategy=_fine_strategy,
        )

        if _fine_strategy not in ("jump", "extend"):
            ve = f"invalid `_fine_strategy`; got {_fine_strategy}"
            raise ValueError(ve)

        self._shape_at = partial(
            coarse2fine_shape,
            self.shape0,
            _coarse_size=_coarse_size,
            _fine_size=_fine_size,
            _fine_strategy=_fine_strategy,
        )

        self._coarse_size = int(_coarse_size)
        self._fine_size = int(_fine_size)
        self._fine_strategy = _fine_strategy

        # Derived attributes
        self._ndim = len(self.shape)
        self._size = np.prod(self.shape, dtype=int)

        if rg2cart is None and cart2rg is None:
            if distances0 is None and distances is None:
                distances = jnp.ones((self.ndim,))
                distances0 = fine2coarse_distances(
                    distances,
                    depth,
                    _fine_size=_fine_size,
                    _fine_strategy=_fine_strategy,
                )
            elif distances0 is not None:
                distances0 = jnp.broadcast_to(jnp.atleast_1d(distances0), (self.ndim,))
                distances = coarse2fine_distances(
                    distances0,
                    depth,
                    _fine_size=_fine_size,
                    _fine_strategy=_fine_strategy,
                )
            else:
                distances = jnp.broadcast_to(jnp.atleast_1d(distances), (self.ndim,))
                distances0 = fine2coarse_distances(
                    distances,
                    depth,
                    _fine_size=_fine_size,
                    _fine_strategy=_fine_strategy,
                )

            def _rg2cart(x):
                x = jnp.asarray(x)
                return x * distances0.reshape((-1,) + (1,) * (x.ndim - 1))

            def _cart2rg(x):
                x = jnp.asarray(x)
                return x / distances0.reshape((-1,) + (1,) * (x.ndim - 1))

            if regular_axes is None and irregular_axes is None:
                regular_axes = tuple(range(self.ndim))
            self._rg2cart = _rg2cart
            self._cart2rg = _cart2rg
        elif rg2cart is not None and cart2rg is not None:
            c0 = jnp.mgrid[tuple(slice(s) for s in self.shape0)]
            if not all(jnp.allclose(r, c) for r, c in zip(cart2rg(rg2cart(c0)), c0)):
                raise ValueError("`cart2rg` is not the inverse of `rg2cart`")
            self._rg2cart = rg2cart
            self._cart2rg = cart2rg
            distances = distances0 = None
        else:
            ve = "invalid combination of `cart2rg`, `rg2cart` and `distances`"
            raise ValueError(ve)
        self.distances = distances
        self.distances0 = distances0

        self.distances_at = partial(
            coarse2fine_distances,
            self.distances0,
            _fine_size=_fine_size,
            _fine_strategy=_fine_strategy,
        )

        if regular_axes is None and irregular_axes is not None:
            regular_axes = tuple(set(range(self.ndim)) - set(irregular_axes))
        elif regular_axes is not None and irregular_axes is None:
            irregular_axes = tuple(set(range(self.ndim)) - set(regular_axes))
        elif regular_axes is None and irregular_axes is None:
            regular_axes = ()
            irregular_axes = tuple(range(self.ndim))
        else:
            if set(regular_axes) | set(irregular_axes) != set(range(self.ndim)):
                ve = "`regular_axes` and `irregular_axes` do not span the full axes"
                raise ValueError(ve)
            if set(regular_axes) & set(irregular_axes) != set():
                ve = "`regular_axes` and `irregular_axes` must be exclusive"
                raise ValueError(ve)
        self._regular_axes = tuple(regular_axes)
        self._irregular_axes = tuple(irregular_axes)
        if len(self.regular_axes) + len(self.irregular_axes) != self.ndim:
            ve = (
                f"length of regular_axes and irregular_axes"
                f" ({len(self.regular_axes)} + {len(self.irregular_axes)} respectively)"
                f" incompatible with overall dimension {self.ndim}"
            )
            raise ValueError(ve)

        self._descr = {
            "depth": self.depth,
            "shape0": self.shape0,
            "_coarse_size": self.coarse_size,
            "_fine_size": self.fine_size,
            "_fine_strategy": self.fine_strategy,
        }
        if distances0 is not None:
            self._descr["distances0"] = tuple(distances0)
        else:
            self._descr["rg2cart"] = repr(rg2cart)
            self._descr["cart2rg"] = repr(cart2rg)
        self._descr["regular_axes"] = self.regular_axes

    @property
    def shape(self):
        """Shape at the final refinement level"""
        return self._shape

    @property
    def shape0(self):
        """Shape at the zeroth refinement level"""
        return self._shape0

    @property
    def size(self):
        return self._size

    @property
    def ndim(self):
        return self._ndim

    @property
    def depth(self):
        return self._depth

    @property
    def coarse_size(self):
        return self._coarse_size

    @property
    def fine_size(self):
        return self._fine_size

    @property
    def fine_strategy(self):
        return self._fine_strategy

    @property
    def regular_axes(self):
        return self._regular_axes

    @property
    def irregular_axes(self):
        return self._irregular_axes

    def rg2cart(self, positions):
        """Translates positions from the regular Euclidean coordinate system to
        the (in general) irregular Cartesian coordinate system.

        Parameters
        ----------
        positions :
            Positions on a regular Euclidean coordinate system.

        Returns
        -------
        positions :
            Positions on an (in general) irregular Cartesian coordinate system.

        Note
        ----
        This method is independent of the refinement level!
        """
        l = len(positions)
        if l != self.ndim:
            ve = f"`positions` of length {l} but chart is {self.ndim}-dimensional"
            raise ValueError(ve)
        return self._rg2cart(positions)

    def cart2rg(self, positions):
        """Translates positions from the (in general) irregular Cartesian
        coordinate system to the regular Euclidean coordinate system.

        Parameters
        ----------
        positions :
            Positions on an (in general) irregular Cartesian coordinate system.

        Returns
        -------
        positions :
            Positions on a regular Euclidean coordinate system.

        Note
        ----
        This method is independent of the refinement level!
        """
        l = len(positions)
        if l != self.ndim:
            ve = f"`positions` of length {l} but chart is {self.ndim}-dimensional"
            raise ValueError(ve)
        return self._cart2rg(positions)

    def rgoffset(self, lvl: int) -> Tuple[float]:
        """Calculate the offset on the regular Euclidean grid due to shrinking
        of the grid with increasing refinement level.

        Parameters
        ----------
        lvl :
            Level of the refinement.

        Returns
        -------
        offset :
            The offset on the regular Euclidean grid along each axes.

        Note
        ----
        Indices are assumed to denote the center of the pixels, i.e. the pixel
        with index `0` is assumed to be at `(0., ) * ndim`.
        """
        csz = self.coarse_size  # abbreviations for readability
        fsz = self.fine_size

        leftmost_center = 0.0
        # Assume the indices denote the center of the pixels, i.e. the pixel
        # with index 0 is at (0., ) * ndim
        if self.fine_strategy == "jump":
            # for i in range(lvl):
            #     leftmost_center += ((csz - 1) / 2 - 0.5 + 0.5 / fsz) / fsz**i
            lm0 = (csz - 1) / 2 - 0.5 + 0.5 / fsz
            geo = (1.0 - fsz**-lvl) / (
                1.0 - 1.0 / fsz
            )  # sum(fsz**-i for i in range(lvl))
            leftmost_center = lm0 * geo
        elif self.fine_strategy == "extend":
            # for i in range(lvl):
            #     leftmost_center += ((csz - 1) / 2 - 0.25 * (fsz - 1)) / 2**i
            lm0 = (csz - 1) / 2 - 0.25 * (fsz - 1)
            geo = (1.0 - 2.0**-lvl) * 2.0  # sum(fsz**-i for i in range(lvl))
            leftmost_center = lm0 * geo
        else:
            raise AssertionError()
        return tuple((leftmost_center,) * self.ndim)

    def ind2rg(self, indices: Iterable[Union[float, int]], lvl: int) -> Tuple[float]:
        """Converts pixel indices to a continuous regular Euclidean grid
        coordinates.

        Parameters
        ----------
        indices :
            Indices of shape `(n_dim, n_indices)` into the NDArray at
            refinement level `lvl` which to convert to points in our regular
            Euclidean grid.
        lvl :
            Level of the refinement.

        Returns
        -------
        rg :
            Regular Euclidean grid coordinates of shape `(n_dim, n_indices)`.
        """
        l = len(indices)
        if l != self.ndim:
            ve = f"`indices` of length {l} but chart is {self.ndim}-dimensional"
            raise ValueError(ve)

        offset = self.rgoffset(lvl)

        if self.fine_strategy == "jump":
            dvol = 1 / self.fine_size**lvl
        elif self.fine_strategy == "extend":
            dvol = 1 / 2**lvl
        else:
            raise AssertionError()
        return tuple(off + idx * dvol for off, idx in zip(offset, indices))

    def rg2ind(
        self, positions: Iterable[Union[float, int]], lvl: int, discretize: bool = True
    ) -> Union[Tuple[float], Tuple[int]]:
        """Converts continuous regular grid positions to pixel indices.

        Parameters
        ----------
        positions :
            Positions on the regular Euclidean coordinate system of shape
            `(n_dim, n_indices)` at refinement level `lvl` which to convert to
            indices in a NDArray at the refinement level `lvl`.
        lvl :
            Level of the refinement.
        discretize :
            Whether to round indices to the next closest integer.

        Returns
        -------
        indices :
            Indices into the NDArray at refinement level `lvl`.
        """
        l = len(positions)
        if l != self.ndim:
            ve = f"`positions` of length {l} but chart is {self.ndim}-dimensional"
            raise ValueError(ve)

        offset = self.rgoffset(lvl)

        if self.fine_strategy == "jump":
            dvol = 1 / self.fine_size**lvl
        elif self.fine_strategy == "extend":
            dvol = 1 / 2**lvl
        else:
            raise AssertionError()
        indices = tuple((pos - off) / dvol for off, pos in zip(offset, positions))
        if discretize:
            indices = tuple(jnp.rint(idx).astype(jnp.int32) for idx in indices)
        return indices

    def ind2cart(self, indices: Iterable[Union[float, int]], lvl: int):
        """Computes the Cartesian coordinates of a pixel given the indices of
        it.

        Parameters
        ----------
        indices :
            Indices of shape `(n_dim, n_indices)` into the NDArray at
            refinement level `lvl` which to convert to locations in our (in
            general) irregular coordinate system of the modeled points.
        lvl :
            Level of the refinement.

        Returns
        -------
        positions :
            Positions in the (in general) irregular coordinate system of the
            modeled points of shape `(n_dim, n_indices)`.
        """
        return self.rg2cart(self.ind2rg(indices, lvl))

    def cart2ind(self, positions, lvl, discretize=True):
        """Computes the indices of a pixel given the Cartesian coordinates of
        it.

        Parameters
        ----------
        positions :
            Positions on the Cartesian (in general) irregular coordinate system
            of the modeled points of shape `(n_dim, n_indices)` at refinement
            level `lvl` which to convert to indices in a NDArray at the
            refinement level `lvl`.
        lvl :
            Level of the refinement.
        discretize :
            Whether to round indices to the next closest integer.

        Returns
        -------
        indices :
            Indices into the NDArray at refinement level `lvl`.
        """
        return self.rg2ind(self.cart2rg(positions), lvl, discretize=discretize)

    def shape_at(self, lvl):
        """Retrieves the shape at a given refinement level `lvl`."""
        return self._shape_at(lvl)

    def level_of(self, shape: Tuple[int]):
        """Finds the refinement level at which the number of grid points
        equate.
        """
        if not isinstance(shape, tuple):
            raise TypeError(f"invalid type of `shape`; got {type(shape)}")

        for lvl in range(self.depth + 1):
            if shape == self.shape_at(lvl):
                return lvl
        else:
            raise ValueError(f"invalid shape {shape!r}")

    def __repr__(self):
        return f"{self.__class__.__name__}(**{self._descr})"

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __hash__(self):
        return hash(repr(self))


def _is_integer(maybe_int):
    return np.asarray(maybe_int, dtype=float).item().is_integer()


class HEALPixChart:
    def __init__(
        self,
        *,
        min_shape: Optional[Iterable[int]],
        depth: int = -1,
        shape0: Optional[Iterable[int]] = None,
        nonhp_rg2cart: Optional[
            Callable[
                [
                    Iterable,
                ],
                Iterable,
            ]
        ],
        nonhp_cart2rg: Optional[
            Callable[
                [
                    Iterable,
                ],
                Iterable,
            ]
        ],
        _coarse_size: int = 3,
        _fine_size: int = 2,
        _fine_strategy: Literal["jump", "extend"] = "extend",
        regular_axes: Optional[Union[Iterable[int], Tuple]] = None,
        irregular_axes: Optional[Union[Iterable[int], Tuple]] = None,
    ):
        """Initialize a refinement chart with HEALPix pixelization on the first
        axis, see NIFTy's `CoordinateChart` method.

        Parameters
        ----------
        """
        from healpy import pixelfunc
        from .healpix_refine import get_1st_hp_nbrs_idx

        nside, nside0 = None, None
        if min_shape is not None:
            nside = sqrt(min_shape[0] / 12)
        elif shape0 is not None:
            nside0 = sqrt(shape0[0] / 12)
        else:
            raise ValueError("one of `min_shape` or `shape0` must be specified")

        if depth < 0:
            if not nside:
                raise ValueError("need `min_shape` to compute `depth`")
            depth = log2(nside)
        if not _is_integer(depth):
            raise ValueError(f"`depth` ({depth!r}) must be an integer")
        self._depth = int(depth)

        if nside is not None:
            nside0 = nside / 2**self.depth
        elif nside0 is not None:
            nside = nside0 * 2**self.depth
        else:
            raise TypeError("specify one of `nside` or `nside0`")
        if not _is_integer(nside) or not _is_integer(nside0):
            ve = f"`nside{{,0}}` must be a power of 2; got ({nside!r}, {nside0!r})"
            raise ValueError(ve)
        self._nside = int(nside)
        self._nside0 = int(nside0)

        if _fine_strategy not in ("jump", "extend"):
            ve = f"invalid `_fine_strategy`; got {_fine_strategy}"
            raise ValueError(ve)
        self._coarse_size = int(_coarse_size)
        self._fine_size = int(_fine_size)
        self._fine_strategy = _fine_strategy

        min_shape = (min_shape,) if isinstance(min_shape, int) else min_shape
        if shape0 is None and min_shape is not None:
            nonhp_shape0 = fine2coarse_shape(
                min_shape[1:],
                self.depth,
                ceil_sizes=True,
                _coarse_size=_coarse_size,
                _fine_size=_fine_size,
                _fine_strategy=_fine_strategy,
            )
            shape0 = (12 * self.nside0**2,) + nonhp_shape0
        elif shape0 is None:
            raise ValueError("either `shape0` or `min_shape` must be specified")
        self._shape0 = shape0
        self._shape = self.shape_at(self.depth)

        # Derived attributes
        self._ndim = len(self.shape)
        self._size = np.prod(self.shape, dtype=int)

        c0 = np.mgrid[tuple(slice(s) for s in self.shape0[1:])]
        if nonhp_cart2rg is not None and nonhp_rg2cart is not None:
            if not all(
                np.allclose(r, c) for r, c in zip(nonhp_cart2rg(nonhp_rg2cart(c0)), c0)
            ):
                raise ValueError(
                    "`nonhp_cart2rg` is not the inverse of `nonhp_rg2cart`"
                )
        self._nonhp_rg2cart = nonhp_rg2cart
        self._nonhp_cart2rg = nonhp_cart2rg

        if regular_axes is None and irregular_axes is not None:
            regular_axes = tuple(set(range(self.ndim)) - set(irregular_axes))
        elif regular_axes is not None and irregular_axes is None:
            irregular_axes = tuple(set(range(self.ndim)) - set(regular_axes))
        elif regular_axes is None and irregular_axes is None:
            regular_axes = ()
            irregular_axes = tuple(range(self.ndim))
        self._regular_axes = tuple(regular_axes)
        self._irregular_axes = tuple(irregular_axes)
        if set(self.regular_axes) | set(self.irregular_axes) != set(range(self.ndim)):
            ve = "`regular_axes` and `irregular_axes` do not span the full axes"
            raise ValueError(ve)
        if set(self.regular_axes) & set(self.irregular_axes) != set():
            ve = "`regular_axes` and `irregular_axes` must be exclusive"
            raise ValueError(ve)

        if 0 not in self.irregular_axes:
            raise ValueError("zeroth HEALPix axis must be irregular")

        if self.coarse_size != 3 or self.fine_size != 2:
            nie = "only `3→2` radial chart is currently supported"
            raise NotImplementedError(nie)

        self.nest = NEST
        self._hp_neighbors_idx = {}
        self._hp_neighbors = {}
        self._hp_children = {}
        for lvl in range(self.depth):
            nside = self.nside_at(lvl)
            pix_idx = np.arange(12 * nside**2)
            self._hp_neighbors_idx[lvl] = get_1st_hp_nbrs_idx(nside, pix_idx, nest=NEST)
            self._hp_neighbors[lvl] = np.stack(
                pixelfunc.pix2vec(nside, self._hp_neighbors_idx[lvl], nest=NEST),
                axis=-1,
            )
            i = pixelfunc.ring2nest(nside, pix_idx) if NEST is False else pix_idx
            self._hp_children[lvl] = np.stack(
                pixelfunc.pix2vec(
                    2 * nside, 4 * i[:, None] + np.arange(0, 4)[None, :], nest=True
                ),
                axis=-1,
            )
            # Cast to JAX arrays to allow jit-able indexing
            self._hp_neighbors_idx[lvl] = jnp.array(self._hp_neighbors_idx[lvl])
            self._hp_neighbors[lvl] = jnp.array(self._hp_neighbors[lvl])
            self._hp_children[lvl] = jnp.array(self._hp_children[lvl])

        self._descr = {
            "depth": self.depth,
            "shape0": self.shape0,
            "_coarse_size": self.coarse_size,
            "_fine_size": self.fine_size,
            "_fine_strategy": self.fine_strategy,
        }
        self._descr["nonhp_rg2cart"] = repr(nonhp_rg2cart)
        self._descr["nonhp_cart2rg"] = repr(nonhp_cart2rg)
        self._descr["regular_axes"] = self.regular_axes

    @property
    def nside(self) -> int:
        return self._nside

    @property
    def nside0(self) -> int:
        return self._nside0

    @property
    def shape(self):
        """Shape at the final refinement level"""
        return self._shape

    @property
    def shape0(self):
        """Shape at the zeroth refinement level"""
        return self._shape0

    @property
    def size(self):
        return self._size

    @property
    def ndim(self):
        return self._ndim

    @property
    def depth(self):
        return self._depth

    @property
    def coarse_size(self):
        return self._coarse_size

    @property
    def fine_size(self):
        return self._fine_size

    @property
    def fine_strategy(self):
        return self._fine_strategy

    @property
    def regular_axes(self):
        return self._regular_axes

    @property
    def irregular_axes(self):
        return self._irregular_axes

    def nside_at(self, lvl: int):
        return self.nside0 * 2**lvl

    def shape_at(self, lvl):
        nonhp_shape = coarse2fine_shape(
            self.shape0[1:],
            lvl,
            _coarse_size=self.coarse_size,
            _fine_size=self.fine_size,
            _fine_strategy=self.fine_strategy,
        )
        return (12 * self.nside_at(lvl) ** 2,) + nonhp_shape

    def rgoffset(self, lvl: int) -> Tuple[float]:
        csz = self.coarse_size  # abbreviations for readability
        fsz = self.fine_size

        leftmost_center = 0.0
        # Assume the indices denote the center of the pixels, i.e. the pixel
        # with index 0 is at (0., ) * ndim
        if self.fine_strategy == "jump":
            # for i in range(lvl):
            #     leftmost_center += ((csz - 1) / 2 - 0.5 + 0.5 / fsz) / fsz**i
            lm0 = (csz - 1) / 2 - 0.5 + 0.5 / fsz
            geo = (1.0 - fsz**-lvl) / (
                1.0 - 1.0 / fsz
            )  # sum(fsz**-i for i in range(lvl))
            leftmost_center = lm0 * geo
        elif self.fine_strategy == "extend":
            # for i in range(lvl):
            #     leftmost_center += ((csz - 1) / 2 - 0.25 * (fsz - 1)) / 2**i
            lm0 = (csz - 1) / 2 - 0.25 * (fsz - 1)
            geo = (1.0 - 2.0**-lvl) * 2.0  # sum(fsz**-i for i in range(lvl))
            leftmost_center = lm0 * geo
        else:
            raise AssertionError()
        return (0.0,) + tuple((leftmost_center,) * (self.ndim - 1))

    rgoffset.__doc__ = CoordinateChart.rgoffset.__doc__

    def nonhp_ind2cart(self, indices: Iterable[int], lvl: int) -> Tuple[float]:
        l = len(indices)
        if l != self.ndim - 1:
            ve = f"non-hp `indices` of length {l} but chart is {self.ndim}-dimensional"
            raise ValueError(ve)

        offset = self.rgoffset(lvl)[1:]

        if self.fine_strategy == "jump":
            dvol = 1 / self.fine_size**lvl
        elif self.fine_strategy == "extend":
            dvol = 1 / 2**lvl
        else:
            raise AssertionError()
        rg1 = tuple(off + idx * dvol for off, idx in zip(offset, indices))
        return self._nonhp_rg2cart(rg1)

    nonhp_ind2cart.__doc__ = CoordinateChart.ind2cart.__doc__

    def hp_neighbors_idx(self, lvl, idx):
        return self._hp_neighbors_idx[lvl][idx]

    def get_coarse_fine_pair(self, indices, lvl: int):
        if self.ndim == 1:
            (i,) = indices
            return self._hp_neighbors[lvl][i], self._hp_children[lvl][i]
        l = len(indices)
        if l != self.ndim:
            ve = f"non-hp `indices` of length {l} but chart is {self.ndim}-dimensional"
            raise ValueError(ve)

        idx_hp, idx_r, *idx_add = indices
        if len(idx_add) > 0:
            raise NotImplementedError()
        ciac = np.arange(self.coarse_size)
        fiac = (np.arange(self.fine_size) - (self.fine_size - 1) / 2) / 2
        fiac += (self.coarse_size - 1) // 2

        gc, gf = self._hp_neighbors[lvl][idx_hp], self._hp_children[lvl][idx_hp]

        bc = (1,) * (self.ndim - 1) + (-1, 1)
        rc = jnp.array(self.nonhp_ind2cart((idx_r + ciac,), lvl)).reshape(bc)
        gc = gc[:, np.newaxis, :] * rc
        gc = gc.reshape(-1, self.ndim + 1)
        rf = jnp.array(self.nonhp_ind2cart((idx_r + fiac,), lvl)).reshape(bc)
        gf = gf[:, np.newaxis, :] * rf
        gf = gf.reshape(-1, self.ndim + 1)

        return gc, gf

    def __repr__(self):
        return f"{self.__class__.__name__}(**{self._descr})"

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __hash__(self):
        return hash(repr(self))



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import dataclasses
from functools import partial

import jax
from jax import numpy as jnp

from ..tree_math import ShapeWithDtype
from ..model import Model


def _los(x, /, start, end, *, distances, shape, n_sampling_points, order=1):
    from jax.scipy.ndimage import map_coordinates

    l2i = ((shape - 1) / shape) / distances
    start_iloc = start * l2i
    end_iloc = end * l2i
    ddi = (end_iloc - start_iloc) / n_sampling_points
    adi = jnp.arange(0, n_sampling_points) + 0.5
    dist = jnp.linalg.norm(end - start)
    pp = start_iloc[:, jnp.newaxis] + ddi[:, jnp.newaxis] * adi[jnp.newaxis]
    return map_coordinates(x, pp, order=order, cval=jnp.nan).sum() * (
        dist / n_sampling_points
    )


class SamplingCartesianGridLOS(Model):
    start: jax.Array = dataclasses.field(metadata=dict(static=False))
    end: jax.Array = dataclasses.field(metadata=dict(static=False))
    distances: jax.Array = dataclasses.field(metadata=dict(static=False))

    def __init__(
        self,
        start,
        end,
        *,
        shape,
        distances,
        n_sampling_points=500,
        interpolation_order=1,
        dtype=None,
    ):
        """Sampling Line-Of-Sight (LOS) intergrator.

        Samples the LOS at a number of points and sum up the result to estimate
        the integral from a starting point to an end point in n-dimensional space.

        Parameters
        ----------
        start :
            Location of the start point(s) in Cartesian space of shape
            `(n_points, n_dim)` or `(n_dim,)`.
        end :
            Location of the end point(s) in Cartesian space of shape
            `(n_points, n_dim)` or `(n_dim,)`.
        shape :
            Shape of the input.
        distances :
            Tuple of distances for each axis of the shape of the input
        n_sampling_points : int, optional
            Number of sampling points per LOS for the integration.
        interpolation_order : int, optional
            Order of the interpolation for reading out the sampling points.
        dtype : data-type, optional
            Hint specifying the dtype for the construction of the domain.
        """
        # We assume that `start` and `end` are of shape (n_points, n_dimensions)
        self.start = jnp.array(start)
        self.end = jnp.array(end)
        self.distances = jnp.array(distances)
        self._los = partial(
            _los,
            n_sampling_points=n_sampling_points,
            order=interpolation_order,
            distances=self.distances,
            shape=jnp.array(shape),
        )
        super().__init__(
            domain=ShapeWithDtype(shape, dtype), target=ShapeWithDtype(end.shape, dtype)
        )

    def __call__(self, x):
        in_axes = (None, 0, 0)
        if self.start.ndim < self.end.ndim:
            in_axes = (None, None, 0)
        elif self.start.ndim > self.end.ndim:
            in_axes = (None, 0, None)
        return jax.vmap(self._los, in_axes=in_axes)(x, self.start, self.end)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from .energy import Energy
from .minimizer import Minimizer


class ADVIOptimizer(Minimizer):
    """Provide an implementation of an adaptive step-size sequence optimizer,
    following https://arxiv.org/abs/1603.00788.

    This stochastic optimizer keeps track of the evolution of the gradient over
    the last steps to adaptively determine the step-size of the next update.
    It is a variation of the Adam optimizer for Gaussian variational inference
    and it allows to optimizer stochastic loss functions.

    Parameters
    ----------
    steps: int
        The number of concecutive steps during one call of the optimizer.
    eta: positive float
        The scale of the step-size sequence. It might have to be adapted to the
        application to increase performance. Default: 1.
    alpha: float between 0 and 1
        The fraction of how much the current gradient impacts the momentum.
        Lower values correspond to a longer memory.
    tau: positive float
        This quantity prevents division by zero.
    epsilon: positive float
        A small value guarantees Robbins and Monro conditions.
    resample: bool
        Whether the loss function is resampled for the next iteration.
        Stochastic losses require resampleing, deterministic ones not.
    """

    def __init__(self, controller, eta=1, alpha=0.1, tau=1, epsilon=1e-16, resample=True):
        self.alpha = alpha
        self.eta = eta
        self.tau = tau
        self.epsilon = epsilon
        self.counter = 1
        self._controller = controller
        self.s = None
        self.resample = resample

    def _step(self, position, gradient):
        self.s = self.alpha * gradient ** 2 + (1 - self.alpha) * self.s
        self.rho = self.eta * self.counter ** (-0.5 + self.epsilon) \
                / (self.tau + (self.s).sqrt())
        new_position = position - self.rho * gradient
        self.counter += 1
        return new_position

    def __call__(self, energy):
        from ..utilities import myassert

        controller = self._controller
        status = controller.start(energy)
        if status != controller.CONTINUE:
            return energy, status

        if self.s is None:
            self.s = energy.gradient ** 2
        while True:
            # check if position is at a flat point
            if energy.gradient_norm == 0:
                return energy, controller.CONVERGED

            x = self._step(energy.position, energy.gradient)
            if self.resample:
                energy = energy.resample_at(x)
            myassert(isinstance(energy, Energy))
            myassert(x.domain is energy.position.domain)

            energy = energy.at(x)
            status = self._controller.check(energy)
            if status != controller.CONTINUE:
                return energy, status

    def reset(self):
        self.counter = 1
        self.s = None



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause
# Authors: Gordian Edenhofer, Philipp Frank,
# Matteo Guardiani, Julian Rüstig

import operator
import dataclasses
from collections import namedtuple
from collections.abc import Mapping
from functools import partial, reduce
from typing import Callable, Optional, Tuple, Union

import numpy as np
from jax import numpy as jnp, vmap

from ..config import _config
from .gauss_markov import IntegratedWienerProcess
from .logger import logger
from .misc import wrap
from .model import Model, WrappedCall
from .num import lognormal_prior, normal_prior
from .tree_math import ShapeWithDtype, random_like


def hartley(p, axes=None):
    from jax.numpy import fft

    tmp = fft.fftn(p, axes=axes)
    c = _config.get("hartley_convention")
    add_or_sub = operator.add if c == "non_canonical_hartley" else operator.sub
    return add_or_sub(tmp.real, tmp.imag)


def get_sht(nside, axis, lmax, mmax, nthreads):
    from jaxbind.contrib.jaxducc0 import get_healpix_sht

    jsht = get_healpix_sht(nside, lmax, mmax, 0, nthreads)
    axis = int(axis)

    def f(inp):
        # Explicitly move axes around, may introduce unnecessary copies
        def trafo(x):
            return np.sqrt(4 * np.pi) * jsht(x[jnp.newaxis])[0][0]

        axs = axis % inp.ndim
        for i in reversed(range(inp.ndim)):
            if i < axs:
                trafo = vmap(trafo, in_axes=0, out_axes=0)
            elif i > axs:
                trafo = vmap(trafo, in_axes=1, out_axes=1)
        return trafo(inp)

    return f


def _unique_mode_distributor(m_length, uniqueness_rtol=1e-12):
    # Construct an array of unique mode lengths
    um = np.unique(m_length)
    tol = uniqueness_rtol * um[-1]
    um = um[np.diff(np.append(um, 2 * um[-1])) > tol]
    # Group modes based on their length and store the result as power
    # distributor
    binbounds = 0.5 * (um[:-1] + um[1:])
    m_length_idx = np.searchsorted(binbounds, m_length)
    m_count = np.bincount(m_length_idx.ravel(), minlength=um.size)
    if np.any(m_count == 0) or um.shape != m_count.shape:
        raise RuntimeError("invalid harmonic mode(s) encountered")
    return m_length_idx, um, m_count


def get_spherical_mode_distributor(
    nside: int, lmax=None, mmax=None, uniqueness_rtol=1e-12, distance_dtype=np.float64
):
    """Get the unique lengths of Spherical harmonic modes, a mapping from a mode
    to its length index and the multiplicity of each unique mode length.

    Parameters
    ----------
    nside : int
        Nside of the HEALPix sphere for which the associated harmonic space is
        constructed
    lmax : int, optional
        The maximum :math:`l` value of any spherical harmonic coefficient
        :math:`a_{lm}` that is represented by this object.
        Must be :math:`\\ge 0`. If not supplied, it is set to `nside * 2`
    mmax : int, optional
        The maximum :math:`m` value of any spherical harmonic coefficient
        :math:`a_{lm}` that is represented by this object.
        If not supplied, it is set to `lmax`.
        Must be :math:`\\ge 0` and :math:`\\le` `lmax`.
    uniqueness_rtol : float
        Relative tolerance to define unique lengths of harmonic mode vectors
        (k-vectors). Vectors with lengths that have a smaller relative distance
        to each other are treated identically
    distance_dtype : Any
        Dtype of the array of harmonic mode lengths canstructed internally

    Returns
    -------
    mode_length_idx : jnp.ndarray
        Index in power-space for every mode in harmonic-space. Can be used to
        distribute power from a power-space to the full harmonic grid.
    unique_mode_length : jnp.ndarray
        Unique length of spherical modes.
    mode_multiplicity : jnp.ndarray
        Multiplicity for each unique mode length.
    """
    if lmax is None:
        lmax = 2 * nside
    lmax = int(lmax)
    if lmax < 0:
        raise ValueError("lmax must be >=0.")
    if mmax is None:
        mmax = lmax
    mmax = int(mmax)
    if mmax < 0 or mmax > lmax:
        raise ValueError("mmax must be >=0 and <=lmax.")
    size = (lmax + 1) ** 2 - (lmax - mmax) * (lmax - mmax + 1)

    ldist = np.empty((size,), dtype=distance_dtype)
    ldist[0 : lmax + 1] = np.arange(lmax + 1, dtype=distance_dtype)
    tmp = np.repeat(np.arange(lmax + 1, dtype=distance_dtype), 2)
    idx = lmax + 1
    for m in range(1, mmax + 1):
        ldist[idx : idx + 2 * (lmax + 1 - m)] = tmp[2 * m :]
        idx += 2 * (lmax + 1 - m)

    return _unique_mode_distributor(ldist, uniqueness_rtol=uniqueness_rtol), (
        lmax,
        mmax,
        size,
    )


def get_fourier_mode_distributor(
    shape: Union[tuple, int], distances: Union[tuple, float], uniqueness_rtol=1e-12
):
    """Get the unique lengths of the Fourier modes, a mapping from a mode to
    its length index and the multiplicity of each unique Fourier mode length.

    Parameters
    ----------
    shape : tuple of int or int
        Position-space shape.
    distances : tuple of float or float
        Position-space distances.

    Returns
    -------
    mode_length_idx : jnp.ndarray
        Index in power-space for every mode in harmonic-space. Can be used to
        distribute power from a power-space to the full harmonic grid.
    unique_mode_length : jnp.ndarray
        Unique length of Fourier modes.
    mode_multiplicity : jnp.ndarray
        Multiplicity for each unique Fourier mode length.
    uniqueness_rtol : float
        Relative tolerance to define unique lengths of harmonic mode vectors
        (k-vectors). Vectors with lengths that have a smaller relative distance
        to each other are treated identically
    """
    shape = (shape,) if isinstance(shape, int) else tuple(shape)

    # Compute length of modes
    mspc_distances = 1.0 / (np.array(shape) * np.array(distances))
    m_length = np.arange(shape[0])
    m_length = np.minimum(m_length, shape[0] - m_length) * mspc_distances[0]
    if len(shape) != 1:
        m_length *= m_length
        for i in range(1, len(shape)):
            tmp = np.arange(shape[i])
            tmp = np.minimum(tmp, shape[i] - tmp) * mspc_distances[i]
            tmp *= tmp
            m_length = np.expand_dims(m_length, axis=-1) + tmp
        m_length = np.sqrt(m_length)

    return _unique_mode_distributor(m_length, uniqueness_rtol=uniqueness_rtol)


RegularCartesianGrid = namedtuple(
    "RegularCartesianGrid",
    (
        "shape",
        "total_volume",
        "distances",
        "harmonic_grid",
    ),
    defaults=(None,),
)

RegularFourierGrid = namedtuple(
    "RegularFourierGrid",
    (
        "shape",
        "power_distributor",
        "mode_multiplicity",
        "mode_lengths",
        "relative_log_mode_lengths",
        "log_volume",
    ),
)

HEALPixGrid = namedtuple(
    "HEALPixGrid",
    (
        "nside",
        "shape",
        "total_volume",
        "harmonic_grid",
    ),
    defaults=(None,),
)

LMGrid = namedtuple(
    "LMGrid",
    (
        "lmax",
        "mmax",
        "shape",
        "power_distributor",
        "mode_multiplicity",
        "mode_lengths",
        "relative_log_mode_lengths",
        "log_volume",
    ),
)


def _log_modes(m_length):
    um = m_length.copy()
    um[1:] = np.log(um[1:])
    um[1:] -= um[1]
    assert um[0] == 0.0
    log_vol = um[2:] - um[1:-1]
    assert um.shape[0] - 2 == log_vol.shape[0]
    return um, log_vol


def make_grid(
    shape, distances, harmonic_type
) -> Union[RegularCartesianGrid, HEALPixGrid]:
    """Creates the grid for the amplitude model"""
    shape = (shape,) if isinstance(shape, int) else tuple(shape)

    # Pre-compute lengths of modes and indices for distributing power
    if harmonic_type.lower() == "fourier":
        distances = tuple(np.broadcast_to(distances, np.shape(shape)))

        totvol = np.prod(np.array(shape) * np.array(distances))
        m_length_idx, m_length, m_count = get_fourier_mode_distributor(shape, distances)
        um, log_vol = _log_modes(m_length)
        harmonic_grid = RegularFourierGrid(
            shape=shape,
            power_distributor=m_length_idx,
            mode_multiplicity=m_count,
            mode_lengths=m_length,
            relative_log_mode_lengths=um,
            log_volume=log_vol,
        )
        # TODO: cache results such that only references are used afterwards
        grid = RegularCartesianGrid(
            shape=shape,
            total_volume=totvol,
            distances=distances,
            harmonic_grid=harmonic_grid,
        )
    elif harmonic_type.lower() == "spherical":
        if len(shape) != 1:
            msg = "`shape` must be length one. Its the nside of the spherical grid."
            raise ValueError(msg)
        nside = shape[0]
        ((m_length_idx, m_length, m_count), (lmax, mmax, size)) = (
            get_spherical_mode_distributor(nside)
        )
        um, log_vol = _log_modes(m_length)
        harmonic_grid = LMGrid(
            lmax=lmax,
            mmax=mmax,
            shape=(size,),
            power_distributor=m_length_idx,
            mode_multiplicity=m_count,
            mode_lengths=m_length,
            relative_log_mode_lengths=um,
            log_volume=log_vol,
        )
        grid = HEALPixGrid(
            nside=nside,
            shape=(12 * nside**2,),
            total_volume=4 * np.pi,
            harmonic_grid=harmonic_grid,
        )
    else:
        ve = f"invalid `harmonic_type` {harmonic_type!r}"
        raise ValueError(ve)
    return grid


def _remove_slope(rel_log_mode_dist, x):
    sc = rel_log_mode_dist / rel_log_mode_dist[-1]
    return x - x[-1] * sc


class MaternAmplitude(Model):
    cutoff: Callable = dataclasses.field(metadata=dict(static=False))
    loglogslope: Callable = dataclasses.field(metadata=dict(static=False))
    scale: Callable = dataclasses.field(metadata=dict(static=False))

    def __init__(
        self,
        grid: Union[RegularCartesianGrid, RegularFourierGrid, HEALPixGrid, LMGrid],
        scale: Optional[Callable],
        cutoff: Callable,
        loglogslope: Callable,
        renormalize_amplitude: bool,
        prefix: str = "",
        kind: str = "amplitude",
    ):
        """Initializes a model that computes the amplitude of a Matérn-kernel
        power spectrum.

        See
        :class:`nifty.re.correlated_field.CorrelatedFieldMaker.add_fluctuations
        _matern`
        for more details on the parameters.

        See also
        --------
        `Causal, Bayesian, & non-parametric modeling of the SARS-CoV-2 viral
        load vs. patient's age`, Guardiani, Matteo and Frank, Philipp and Kostić,
        Andrija and Edenhofer, Gordian and Roth, Jakob and Uhlmann, Berit and
        Enßlin, Torsten, `<https://arxiv.org/abs/2105.13483>`_
        `<https://doi.org/10.1371/journal.pone.0275011>`_
        """

        self.grid = grid
        self.cutoff = WrappedCall(cutoff, name=prefix + "cutoff")
        self.loglogslope = WrappedCall(loglogslope, name=prefix + "loglogslope")
        self.scale = (
            WrappedCall(scale, name=prefix + "scale") if scale is not None else None
        )

        self.kind = kind.lower()

        supported_kinds = {"amplitude", "power"}
        if self.kind not in supported_kinds:
            raise ValueError(
                f"Invalid kind specified {self.kind!r}, "
                f"supported kinds: {supported_kinds}"
            )

        self.renormalize_amplitude = renormalize_amplitude
        if self.renormalize_amplitude:
            logger.warning("Renormalize amplidude is not yet tested!")

        models = [self.scale, self.cutoff, self.loglogslope]
        domain = reduce(operator.or_, [m.domain for m in models if m is not None])

        super().__init__(domain=domain, white_init=True)

    def __call__(self, primals: Mapping) -> jnp.ndarray:
        if self.scale is None:
            scl = 1.0
        else:
            scl = self.scale(primals)

        ctf = self.cutoff(primals)
        slp = self.loglogslope(primals)

        ln_spectrum = (
            0.25 * slp * jnp.log1p((self.grid.harmonic_grid.mode_lengths / ctf) ** 2)
        )
        spectrum = jnp.exp(ln_spectrum)

        norm = 1.0
        if self.renormalize_amplitude:
            if self.kind == "amplitude":
                norm = jnp.sqrt(
                    jnp.sum(
                        self.grid.harmonic_grid.mode_multiplicity[1:]
                        * spectrum[1:] ** 4
                    )
                )
            elif self.kind == "power":
                norm = jnp.sqrt(
                    jnp.sum(
                        self.grid.harmonic_grid.mode_multiplicity[1:]
                        * spectrum[1:] ** 2
                    )
                )

            norm /= jnp.sqrt(
                self.grid.total_volume
            )  # Due to integral in harmonic space
        spectrum = scl * (jnp.sqrt(self.grid.total_volume) / norm) * spectrum
        spectrum = spectrum.at[0].set(self.grid.total_volume)
        if self.kind.lower() == "power":
            spectrum = jnp.sqrt(spectrum)

        return spectrum


class NonParametricAmplitude(Model):
    loglogavgslope: Callable = dataclasses.field(metadata=dict(static=False))
    fluctuations: Callable = dataclasses.field(metadata=dict(static=False))
    flexibility: Optional[Callable] = dataclasses.field(metadata=dict(static=False))
    asperity: Optional[Callable] = dataclasses.field(metadata=dict(static=False))

    def __init__(
        self,
        grid: Union[RegularCartesianGrid, RegularFourierGrid, HEALPixGrid, LMGrid],
        fluctuations: Optional[Callable],
        loglogavgslope: Callable,
        flexibility: Optional[Callable] = None,
        asperity: Optional[Callable] = None,
        prefix: str = "",
        kind: str = "amplitude",
    ):
        """Initializes a model which computes the amplitude of a non-parametric
        power spectrum.

        See
        :class:`nifty.re.correlated_field.CorrelatedFieldMaker.add_fluctuations`
        for more details on the parameters.

        See also
        --------
        `Variable structures in M87* from space, time and frequency resolved
        interferometry`, Arras, Philipp and Frank, Philipp and Haim, Philipp
        and Knollmüller, Jakob and Leike, Reimar and Reinecke, Martin and
        Enßlin, Torsten, `<https://arxiv.org/abs/2002.05218>`_
        `<http://dx.doi.org/10.1038/s41550-021-01548-0>`_
        """
        self.grid = grid
        log_vol = grid.harmonic_grid.log_volume
        self.kind = kind.lower()

        supported_kinds = {"amplitude", "power"}
        if self.kind not in supported_kinds:
            raise ValueError(
                f"Invalid kind specified {self.kind!r}, "
                f"supported kinds: {supported_kinds}"
            )

        self._loglogavgslope = WrappedCall(
            loglogavgslope, name=prefix + "loglogavgslope", white_init=True
        )
        self.fluctuations = (
            WrappedCall(fluctuations, name=prefix + "fluctuations", white_init=True)
            if fluctuations is not None
            else None
        )
        if flexibility is not None and (log_vol.size > 0):
            flexibility = WrappedCall(
                flexibility, name=prefix + "flexibility", white_init=True
            )
            assert log_vol is not None
            assert (
                self.grid.harmonic_grid.relative_log_mode_lengths.ndim
                == log_vol.ndim
                == 1
            )
            if asperity is not None:
                asperity = WrappedCall(
                    asperity, name=prefix + "asperity", white_init=True
                )
            self._deviations = IntegratedWienerProcess(
                jnp.zeros((2,)),
                flexibility,
                log_vol,
                name=prefix + "spectrum",
                asperity=asperity,
            )
        else:
            self._deviations = None

        models = [
            self.fluctuations,
            self._loglogavgslope,
            self._deviations,
        ]
        domain = reduce(operator.or_, [m.domain for m in models if m is not None])

        super().__init__(domain=domain, white_init=True)

    def __call__(self, primals: Mapping) -> jnp.ndarray:
        mode_multiplicity = self.grid.harmonic_grid.mode_multiplicity
        relative_log_mode_lengths = self.grid.harmonic_grid.relative_log_mode_lengths

        flu = 1.0 if self.fluctuations is None else self.fluctuations(primals)
        slope = self._loglogavgslope(primals)
        slope *= relative_log_mode_lengths
        ln_spectrum = slope

        if self._deviations is not None:
            twolog = self._deviations(primals)
            # Prepend zeromode
            twolog = jnp.concatenate((jnp.zeros((1,)), twolog[:, 0]))
            ln_spectrum += _remove_slope(relative_log_mode_lengths, twolog)

        # Exponentiate and norm the power spectrum
        spectrum = jnp.exp(ln_spectrum)

        # Take the sqrt of the integral of the slope w/o fluctuations and
        # zero-mode while taking into account the multiplicity of each mode
        if self.kind == "amplitude":
            norm = jnp.sqrt(jnp.sum(mode_multiplicity[1:] * spectrum[1:] ** 2))
            norm /= jnp.sqrt(
                self.grid.total_volume
            )  # Due to integral in harmonic space
            amplitude = flu * (jnp.sqrt(self.grid.total_volume) / norm) * spectrum
        elif self.kind == "power":
            norm = jnp.sqrt(jnp.sum(mode_multiplicity[1:] * spectrum[1:]))
            norm /= jnp.sqrt(
                self.grid.total_volume
            )  # Due to integral in harmonic space
            amplitude = (
                flu * (jnp.sqrt(self.grid.total_volume) / norm) * jnp.sqrt(spectrum)
            )
        amplitude = amplitude.at[0].set(self.grid.total_volume)
        return amplitude


class CorrelatedFieldMaker:
    """Construction helper for hierarchical correlated field models.

    The correlated field models are parametrized by creating square roots of
    power spectrum operators ("amplitudes") via calls to
    :func:`add_fluctuations*` that act on the targeted field subgrids.
    During creation of the :class:`CorrelatedFieldMaker`, a global offset from
    zero of the field model can be defined and an operator applying
    fluctuations around this offset is parametrized.

    Creation of the model operator is completed by calling the method
    :func:`finalize`, which returns the configured operator.

    See the method's initialization, :func:`add_fluctuations`,
    :func:`add_fluctuations_matern` and :func:`finalize` for further
    usage information."""

    def __init__(self, prefix: str):
        """Instantiate a CorrelatedFieldMaker object.

        Parameters
        ----------
        prefix : string
            Prefix to the names of the parameters of the CF operator.
        """
        self._azm = None
        self._offset_mean = None
        self._fluctuations = []
        self._target_grids = []
        self._parameter_tree = {}

        self._prefix = prefix

    def add_fluctuations(
        self,
        shape: Union[tuple, int],
        distances: Union[tuple, float],
        fluctuations: Union[tuple, Callable],
        loglogavgslope: Union[tuple, Callable],
        flexibility: Union[tuple, Callable, None] = None,
        asperity: Union[tuple, Callable, None] = None,
        prefix: str = "",
        harmonic_type: str = "fourier",
        non_parametric_kind: str = "amplitude",
    ):
        """Adds a correlation structure to the to-be-made field.

        Correlations are described by their power spectrum and the subgrid on
        which they apply.

        Multiple calls to `add_fluctuations` are possible, in which case
        the constructed field will have the outer product of the individual
        power spectra as its global power spectrum.

        The parameters `fluctuations`, `flexibility`, `asperity` and
        `loglogavgslope` configure either the amplitude or the power
        spectrum model used on the target field subgrid of type
        `harmonic_type`. It is assembled as the sum of a power law component
        (linear slope in log-log amplitude-frequency-space), a smooth varying
        component (integrated Wiener process) and a ragged component
        (un-integrated Wiener process).

        Parameters
        ----------
        shape : tuple of int
            Shape of the position space grid
        distances : tuple of float or float
            Distances in the position space grid
        fluctuations : tuple of float (mean, std) or callable
            Total spectral energy, i.e. amplitude of the fluctuations
            (by default a priori log-normal distributed)
        loglogavgslope : tuple of float (mean, std) or callable
            Power law component exponent
            (by default a priori normal distributed)
        flexibility : tuple of float (mean, std) or callable or None
            Amplitude of the non-power-law power spectrum component
            (by default a priori log-normal distributed)
        asperity : tuple of float (mean, std) or callable or None
            Roughness of the non-power-law power spectrum component; use it to
            accommodate single frequency peak
            (by default a priori log-normal distributed)
        prefix : str
            Prefix of the power spectrum parameter names
        harmonic_type : str
            Description of the harmonic partner domain in which the amplitude
            lives
        non_parametric_kind : str
            If set to `'amplitude'`, the amplitude spectrum is described
            by the correlated field model parameters in the above.
            If set to `'power'`, the power spectrum is described by the
            correlated field model parameters in the above
            (by default set to `'amplitude'`).

        See also
        --------
        `Variable structures in M87* from space, time and frequency resolved
        interferometry`, Arras, Philipp and Frank, Philipp and Haim, Philipp
        and Knollmüller, Jakob and Leike, Reimar and Reinecke, Martin and
        Enßlin, Torsten, `<https://arxiv.org/abs/2002.05218>`_
        `<http://dx.doi.org/10.1038/s41550-021-01548-0>`_
        """
        grid = make_grid(shape, distances, harmonic_type)

        flu = fluctuations
        if isinstance(flu, (tuple, list)):
            flu = lognormal_prior(*flu)
        elif not callable(flu):
            te = f"invalid `fluctuations` specified; got '{type(fluctuations)}'"
            raise TypeError(te)
        slp = loglogavgslope
        if isinstance(slp, (tuple, list)):
            slp = normal_prior(*slp)
        elif not callable(slp):
            te = f"invalid `loglogavgslope` specified; got '{type(loglogavgslope)}'"
            raise TypeError(te)

        flx = flexibility
        if isinstance(flx, (tuple, list)):
            flx = lognormal_prior(*flx)
        elif flx is not None and not callable(flx):
            te = f"invalid `flexibility` specified; got '{type(flexibility)}'"
            raise TypeError(te)
        asp = asperity
        if isinstance(asp, (tuple, list)):
            asp = lognormal_prior(*asp)
        elif asp is not None and not callable(asp):
            te = f"invalid `asperity` specified; got '{type(asperity)}'"
            raise TypeError(te)

        npa = NonParametricAmplitude(
            grid=grid,
            fluctuations=flu,
            loglogavgslope=slp,
            flexibility=flx,
            asperity=asp,
            prefix=self._prefix + prefix,
            kind=non_parametric_kind,
        )
        self._fluctuations.append(npa)
        self._target_grids.append(grid)
        self._parameter_tree.update(npa.domain)

    def add_fluctuations_matern(
        self,
        shape: Union[tuple, int],
        distances: Union[tuple, float],
        scale: Union[tuple, Callable],
        cutoff: Union[tuple, Callable],
        loglogslope: Union[tuple, Callable],
        renormalize_amplitude: bool,
        prefix: str = "",
        harmonic_type: str = "fourier",
        non_parametric_kind: str = "amplitude",
    ):
        """Adds a Matérn-kernel correlation structure to the
        field to be made.

        The Matérn-kernel spectrum is parametrized by

        .. math ::
            A(k) = \\frac{a}{\\left(1 + { \
                \\left(\\frac{|k|}{b}\\right) \
            }^2\\right)^{-c/4}}

        where :math:`a` is called the scale parameter, :math:`b`
        the represents the cutoff mode, and :math:`c` the spectral index
        of the resulting power spectrum.

        Parameters
        ----------
        shape : tuple of int
            Shape of the position space grid.
        distances : tuple of float or float
            Distances in the position space grid.
        scale : tuple of float (mean, std) or callable
            Total spectral energy, i.e. amplitude of the fluctuations
            (by default a priori log-normal distributed).
        cutoff : tuple of float (mean, std) or callable
            Power law component exponent
            (by default a priori normal distributed).
        loglogslope : tuple of float (mean, std) or callable or None
            Amplitude of the non-power-law power spectrum component
            (by default a priori log-normal distributed).
        renormalize_amplitude : bool
            Whether the amplitude of the process should be renormalized to
            ensure that the `scale` parameter relates to the scale of the
            fluctuations along the specified axis.
        prefix : str
            Prefix of the power spectrum parameter names.
        harmonic_type : str
            Description of the harmonic partner domain in which the amplitude
            lives.
        non_parametric_kind : str
            If set to `'amplitude'`, the amplitude spectrum is described
            by the Matérn kernel function in the above.
            If set to `'power'`, the power spectrum is described by the
            Matérn kernel function in the above
            (by default `'amplitude'`).

        See also
        --------
        `Causal, Bayesian, & non-parametric modeling of the SARS-CoV-2 viral
        load vs. patient's age`, Guardiani, Matteo and Frank, Kostić Andrija
        and Edenhofer, Gordian and Roth, Jakob and Uhlmann, Berit and
        Enßlin, Torsten, `<https://arxiv.org/abs/2105.13483>`_
        `<https://doi.org/10.1371/journal.pone.0275011>`_
        """
        grid = make_grid(shape, distances, harmonic_type)

        if isinstance(scale, (tuple, list)):
            scale = lognormal_prior(*scale)
        elif not callable(scale):
            te = f"invalid `scale` specified; got '{type(scale)}'"
            raise TypeError(te)
        if isinstance(cutoff, (tuple, list)):
            cutoff = lognormal_prior(*cutoff)
        elif not callable(cutoff):
            te = f"invalid `cutoff` specified; got '{type(cutoff)}'"
            raise TypeError(te)
        if isinstance(loglogslope, (tuple, list)):
            loglogslope = normal_prior(*loglogslope)
        elif not callable(loglogslope):
            te = f"invalid `loglogslope` specified; got '{type(loglogslope)}'"
            raise TypeError(te)

        ma = MaternAmplitude(
            grid=grid,
            scale=scale,
            cutoff=cutoff,
            loglogslope=loglogslope,
            prefix=self._prefix + prefix,
            kind=non_parametric_kind,
            renormalize_amplitude=renormalize_amplitude,
        )
        self._fluctuations.append(ma)
        self._target_grids.append(grid)
        self._parameter_tree.update(ma.domain)

    def set_amplitude_total_offset(
        self, offset_mean: float, offset_std: Union[tuple, Callable]
    ):
        """Sets the zero-mode for the combined amplitude operator

        Parameters
        ----------
        offset_mean : float
            Mean offset from zero of the correlated field to be made.
        offset_std : tuple of float or callable
            Mean standard deviation and standard deviation of the standard
            deviation of the offset. No, this is not a word duplication.
            (By default a priori log-normal distributed)
        """
        if self._offset_mean is not None and self._azm is not None:
            msg = "Overwriting the previous mean offset and zero-mode"
            logger.warning(msg)

        self._offset_mean = offset_mean
        zm = offset_std
        if not callable(zm):
            if zm is None or len(zm) != 2:
                raise TypeError(f"`offset_std` of invalid type {type(zm)!r}")
            zm = lognormal_prior(*zm)

        self._azm = wrap(zm, self._prefix + "zeromode")
        self._parameter_tree[self._prefix + "zeromode"] = ShapeWithDtype(())

    @property
    def amplitude_total_offset(self) -> Callable:
        """Returns the total offset of the amplitudes"""
        if self._azm is None:
            nie = "You need to set the `amplitude_total_offset` first"
            raise NotImplementedError(nie)
        return self._azm

    @property
    def azm(self):
        """Alias for `amplitude_total_offset`"""
        return self.amplitude_total_offset

    @property
    def fluctuations(self) -> Tuple[Callable, ...]:
        """Returns the added fluctuations, i.e. un-normalized amplitudes

        Their scales are only meaningful relative to one another. Their
        absolute scale bares no information.
        """
        return tuple(self._fluctuations)

    def get_normalized_amplitudes(self) -> Tuple[Callable, ...]:
        """Returns the normalized amplitude operators used in the final model

        The amplitude operators are corrected for the otherwise degenerate
        zero-mode. Their scales are only meaningful relative to one another.
        Their absolute scale bares no information.
        """

        def _mk_normed_amp(amp):  # Avoid late binding
            def normed_amplitude(p):
                return amp(p).at[1:].mul(1.0 / self.azm(p))

            return normed_amplitude

        return tuple(_mk_normed_amp(amp) for amp in self._fluctuations)

    @property
    def amplitude(self) -> Callable:
        """Returns the added fluctuation, i.e. un-normalized amplitude"""
        if len(self._fluctuations) > 1:
            s = (
                "If more than one spectrum is present in the model,"
                " no unique set of amplitudes exist because only the"
                " relative scale is determined."
            )
            raise NotImplementedError(s)
        amp = self._fluctuations[0]

        def ampliude_w_zm(p):
            return amp(p).at[0].mul(self.azm(p))

        return ampliude_w_zm

    @property
    def power_spectrum(self) -> Callable:
        """Returns the power spectrum"""
        amp = self.amplitude

        def power(p):
            return amp(p) ** 2

        return power

    def finalize(self) -> Model:
        """Finishes off the model construction process and returns the
        constructed operator.
        """
        harmonic_transforms = []
        excitation_shape = ()
        for sgrid in self._target_grids:
            sub_shp = None
            sub_shp = sgrid.harmonic_grid.shape
            excitation_shape += sub_shp
            n = len(excitation_shape)
            harmonic_dvol = 1.0 / sgrid.total_volume
            if isinstance(sgrid, RegularCartesianGrid):
                axes = tuple(range(n - len(sub_shp), n))
                # TODO: Generalize to complex
                trafo = partial(hartley, axes=axes)
            elif isinstance(sgrid, HEALPixGrid):
                axis = len(excitation_shape) - 1
                trafo = get_sht(
                    nside=sgrid.nside,
                    axis=axis,
                    lmax=sgrid.harmonic_grid.lmax,
                    mmax=sgrid.harmonic_grid.mmax,
                    nthreads=1,
                )
            harmonic_transforms.append((harmonic_dvol, trafo))

        # Register the parameters for the excitations in harmonic space
        # TODO: actually account for the dtype here
        pfx = self._prefix + "xi"
        self._parameter_tree[pfx] = ShapeWithDtype(excitation_shape)

        def outer_harmonic_transform(p):
            harmonic_dvol, ht = harmonic_transforms[0]
            outer = harmonic_dvol * ht(p)
            for harmonic_dvol, ht in harmonic_transforms[1:]:
                outer = harmonic_dvol * ht(outer)
            return outer

        def _mk_expanded_amp(amp, sub_dom):  # Avoid late binding
            def expanded_amp(p):
                return amp(p)[sub_dom.harmonic_grid.power_distributor]

            return expanded_amp

        expanded_amplitudes = []
        namps = self.get_normalized_amplitudes()
        for amp, sgrid in zip(namps, self._target_grids):
            expanded_amplitudes.append(_mk_expanded_amp(amp, sgrid))

        def outer_amplitude(p):
            outer = expanded_amplitudes[0](p)
            for amp in expanded_amplitudes[1:]:
                # NOTE, the order is important here and must match with the
                # excitations
                # TODO, use functions instead and utilize numpy's casting
                outer = jnp.tensordot(outer, amp(p), axes=0)
            return outer

        def correlated_field(p):
            ea = outer_amplitude(p)
            cf_h = self.azm(p) * ea * p[self._prefix + "xi"]
            return self._offset_mean + outer_harmonic_transform(cf_h)

        init = {
            k: partial(random_like, primals=v) for k, v in self._parameter_tree.items()
        }
        cf = Model(correlated_field, domain=self._parameter_tree.copy(), init=init)
        cf.normalized_amplitudes = namps
        cf.target_grids = tuple(self._target_grids)
        return cf



# Copyright(C) 2013-2021 Gordian Edenhofer
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from .forest_math import (
    assert_arithmetics,
    get_map,
    has_arithmetics,
    map_forest,
    map_forest_mean,
    mean,
    mean_and_std,
    random_like,
    stack,
    tree_shape,
    unstack,
)
from .pytree_string import PyTreeString, hide_strings
from .util import logm, solve, sqrtm
from .vector import Vector
from .vector_math import (
    ShapeWithDtype,
    all,
    any,
    conj,
    conjugate,
    dot,
    matmul,
    max,
    min,
    norm,
    ones_like,
    result_type,
    shape,
    size,
    sum,
    vdot,
    where,
    zeros_like,
)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause
# Author: Philipp Frank, Laurin Soeding, Gordian Edenhofer

from dataclasses import dataclass, field
from typing import Callable, Union

import jax
import jax.numpy as jnp
import numpy as np
from jax.tree_util import Partial, register_dataclass
from numpy import typing as npt
from scipy.special import j0, sici

from ..model import Model
from ..prior import LogNormalPrior, NormalPrior
from ..tree_math import zeros_like

_RP1 = jnp.array(
    [
        -8.99971225705559398224e8,
        4.52228297998194034323e11,
        -7.27494245221818276015e13,
        3.68295732863852883286e15,
    ]
)
_RQ1 = jnp.array(
    [
        1.0,
        6.20836478118054335476e2,
        2.56987256757748830383e5,
        8.35146791431949253037e7,
        2.21511595479792499675e10,
        4.74914122079991414898e12,
        7.84369607876235854894e14,
        8.95222336184627338078e16,
        5.32278620332680085395e18,
    ]
)

_PP1 = jnp.array(
    [
        7.62125616208173112003e-4,
        7.31397056940917570436e-2,
        1.12719608129684925192e0,
        5.11207951146807644818e0,
        8.42404590141772420927e0,
        5.21451598682361504063e0,
        1.00000000000000000254e0,
    ]
)
_PQ1 = jnp.array(
    [
        5.71323128072548699714e-4,
        6.88455908754495404082e-2,
        1.10514232634061696926e0,
        5.07386386128601488557e0,
        8.39985554327604159757e0,
        5.20982848682361821619e0,
        9.99999999999999997461e-1,
    ]
)

_QP1 = jnp.array(
    [
        5.10862594750176621635e-2,
        4.98213872951233449420e0,
        7.58238284132545283818e1,
        3.66779609360150777800e2,
        7.10856304998926107277e2,
        5.97489612400613639965e2,
        2.11688757100572135698e2,
        2.52070205858023719784e1,
    ]
)
_QQ1 = jnp.array(
    [
        1.0,
        7.42373277035675149943e1,
        1.05644886038262816351e3,
        4.98641058337653607651e3,
        9.56231892404756170795e3,
        7.99704160447350683650e3,
        2.82619278517639096600e3,
        3.36093607810698293419e2,
    ]
)

_Z1 = 1.46819706421238932572e1
_Z2 = 4.92184563216946036703e1
_THPIO4 = 2.35619449019234492885  # 3*pi/4
_SQ2OPI = 0.79788456080286535588  # sqrt(2/pi)


def _j1_small(x):
    z = x * x
    w = jnp.polyval(_RP1, z) / jnp.polyval(_RQ1, z)
    w = w * x * (z - _Z1) * (z - _Z2)
    return w


def _j1_large_c(x):
    w = 5.0 / x
    z = w * w
    p = jnp.polyval(_PP1, z) / jnp.polyval(_PQ1, z)
    q = jnp.polyval(_QP1, z) / jnp.polyval(_QQ1, z)
    xn = x - _THPIO4
    p = p * jnp.cos(xn) - w * q * jnp.sin(xn)
    return p * _SQ2OPI / jnp.sqrt(x)


def j1(x):
    """
    Bessel function of order one - using the implementation from CEPHES,
    translated to Jax.

    Pure JAX implementation for Bessel function from
    https://github.com/benjaminpope/sibylla/blob/main/notebooks/bessel_test.ipynb .
    """
    return jnp.sign(x) * jnp.where(
        jnp.abs(x) < 5.0, _j1_small(jnp.abs(x)), _j1_large_c(jnp.abs(x))
    )


@dataclass
class FourierIntegralGrid:
    mode_lengths: jax.Array
    mode_binbounds: jax.Array
    min_dist: float
    max_dist: float
    num: int
    ndim: int
    weights: jax.Array


register_dataclass(
    FourierIntegralGrid,
    data_fields=["mode_lengths", "mode_binbounds", "weights"],
    meta_fields=["num", "min_dist", "max_dist", "ndim"],
)


def make_integral_grid(
    min_dist: float, max_dist: float, num: int, ndim: int, normalize: bool = True
) -> FourierIntegralGrid:
    mode_lengths = np.geomspace(1.0 / max_dist, 1.0 / min_dist, num, endpoint=False)
    mode_lengths = np.insert(mode_lengths, 0, 0.0)
    # Binbounds
    lk = np.log(mode_lengths[1:])
    dlk = (np.log(max_dist) - np.log(min_dist)) / num
    lk = np.append(lk - 0.5 * dlk, lk[-1] + 0.5 * dlk)
    mode_binbounds = np.insert(np.exp(lk), 0, 0.0)
    # norm_weights
    wgt = None
    if normalize:
        if ndim == 1:
            fkr = sici(mode_binbounds * max_dist)[0]
        elif ndim == 2:
            fkr = 1.0 - j0(mode_binbounds * max_dist)
        elif ndim == 3:
            fkr = sici(mode_binbounds * max_dist)[0] - np.sin(mode_binbounds * max_dist)
        else:
            raise NotImplementedError
        wgt = fkr[1:] - fkr[:-1]
        if (ndim == 1) or (ndim == 3):
            wgt *= 2.0 / np.pi
    return FourierIntegralGrid(
        min_dist=min_dist,
        max_dist=max_dist,
        num=num,
        ndim=ndim,
        mode_lengths=mode_lengths,
        mode_binbounds=mode_binbounds,
        weights=wgt,
    )


def spectrum2covariance(
    fig: FourierIntegralGrid, spec: npt.NDArray, *, ref_distance=1.0, normalize=True
):
    fct = [np.pi, 2.0 * np.pi, 2.0 * np.pi**2]

    def cov(r: npt.NDArray) -> npt.NDArray:
        k = jnp.expand_dims(fig.mode_binbounds, tuple(i for i in range(len(r.shape))))
        r = r[..., jnp.newaxis]
        kr = r * k
        if fig.ndim == 1:
            fkr = jnp.sin(kr)
        elif fig.ndim == 2:
            fkr = kr * j1(kr)
        elif fig.ndim == 3:
            fkr = jnp.sin(kr) - kr * jnp.cos(kr)
        else:
            raise NotImplementedError
        res0 = (k[..., 1:] ** fig.ndim - k[..., :-1] ** fig.ndim) / fig.ndim
        resn0 = (fkr[..., 1:] - fkr[..., :-1]) / r**fig.ndim
        res = jnp.where(r < ref_distance * 1e-10, res0, resn0) / fct[fig.ndim - 1]
        res = jnp.tensordot(res, spec, axes=(-1, 0))
        if normalize:
            res /= (fig.weights * spec).sum()
        return res

    return cov


class MaternHarmonicCovariance(Model):
    scale: Union[Model, float] = field(metadata=dict(static=False))
    cutoff: Union[Model, float] = field(metadata=dict(static=False))
    loglogslope: Union[Model, float] = field(metadata=dict(static=False))
    _interp_dists: jnp.ndarray = field(metadata=dict(static=False))
    _fig: FourierIntegralGrid = field(metadata=dict(static=False))

    def __init__(
        self,
        scale: Union[tuple, Callable, float],
        cutoff: Union[tuple, Model, float],
        loglogslope: Union[tuple, Callable, float],
        *,
        ndim: int,
        n_integrate=2_000,
        n_interpolate=512,
        integration_dists_min_max=None,
        interpolation_dists_min_max=None,
        kind: str = "amplitude",
        prefix: str = "",
    ):
        """Compute the Matérn-kernel using its harmonic representation.

        See also
        --------
        `Causal, Bayesian, & non-parametric modeling of the SARS-CoV-2 viral
        load vs. patient's age`, Guardiani, Matteo and Frank, Philipp and Kostić,
        Andrija and Edenhofer, Gordian and Roth, Jakob and Uhlmann, Berit and
        Enßlin, Torsten, `<https://arxiv.org/abs/2105.13483>`_
        `<https://doi.org/10.1371/journal.pone.0275011>`_
        """
        ref_distance = 1.0
        if isinstance(cutoff, (tuple, list)):
            ref_distance *= cutoff[0]
            cutoff = LogNormalPrior(*cutoff, name=prefix + "cutoff")
        elif isinstance(cutoff, Model):
            ref_distance *= cutoff(zeros_like(cutoff.domain))
        elif isinstance(cutoff, float):
            ref_distance *= cutoff
        else:
            raise TypeError(f"invalid `cutoff` specified; got '{cutoff!r}'")
        self.cutoff = cutoff
        self._ref_distance = ref_distance
        if isinstance(loglogslope, (tuple, list)):
            loglogslope = NormalPrior(*loglogslope, name=prefix + "loglogslope")
        elif not (callable(loglogslope) or isinstance(loglogslope, float)):
            raise TypeError(f"invalid `loglogslope` specified; got '{loglogslope!r}'")
        self.loglogslope = loglogslope
        if isinstance(scale, (tuple, list)):
            scale = LogNormalPrior(*scale, name=prefix + "scale")
        elif not (callable(scale) or isinstance(scale, float)):
            raise TypeError(f"invalid `scale` specified; got '{scale!r}'")
        self.scale = scale

        self.kind = kind
        self.ndim = ndim

        if integration_dists_min_max is None:
            integration_dists_min_max = tuple(
                self._ref_distance * np.array([1e-3, 1e4])
            )
        if interpolation_dists_min_max is None:
            interpolation_dists_min_max = tuple(
                self._ref_distance * np.array([1e-3, 1e2])
            )
        self._interp_dists = jnp.geomspace(*interpolation_dists_min_max, n_interpolate)
        min_dist, max_dist = integration_dists_min_max
        self._fig = make_integral_grid(
            min_dist, max_dist, n_integrate, ndim=self.ndim, normalize=False
        )

        super().__init__(
            domain=getattr(self.scale, "domain", {})
            | getattr(self.loglogslope, "domain", {})
            | getattr(self.cutoff, "domain", {})
        )

    def normalized_spectrum(self, x):
        cutoff = self.cutoff(x) if callable(self.cutoff) else self.cutoff
        loglogslope = (
            self.loglogslope(x) if callable(self.loglogslope) else self.loglogslope
        )

        ln_spectrum = (
            0.25 * loglogslope * jnp.log1p((self._fig.mode_lengths / cutoff) ** 2)
        )

        spectrum = jnp.exp(ln_spectrum)
        spectrum = spectrum.at[0].set(spectrum[1])
        if self.kind.lower() == "amplitude":
            spectrum = spectrum**2
        elif self.kind.lower() != "power":
            raise ValueError(f"invalid kind specified {self.kind!r}")
        return spectrum

    @staticmethod
    def _interp_cov(x, y, *, scale, distances, logcorr):
        r = jnp.linalg.norm(x - y, axis=0, ord=2)
        cov = jnp.exp(
            jnp.interp(r, distances, logcorr, left="extrapolate", right="extrapolate")
        )
        cov = jnp.where(r == 0.0, jnp.ones_like(r), cov)
        return scale * cov

    def __call__(self, x):
        scale = self.scale(x) if callable(self.scale) else self.scale
        spec = self.normalized_spectrum(x)
        corr_func = spectrum2covariance(
            self._fig, spec, ref_distance=self._ref_distance, normalize=False
        )

        corr = jax.vmap(corr_func)(self._interp_dists) / corr_func(jnp.array([0.0]))
        # This could perhaps go even lower, but the accuaracy is already quite
        # good and this is definitely numerically stable.
        # Restrict domain to ensure good numerics
        ref_scale = 1e-5
        mask = corr < ref_scale
        maxidx = jnp.argmax(mask) - 1
        # Take abs to ensure good logarithm
        ln_corr = jnp.log(jnp.abs(corr))
        # DIY interpolation that keeps the shapes static
        slope_at_maxidx = (ln_corr[maxidx - 1] - ln_corr[maxidx]) / (
            self._interp_dists[maxidx - 1] - self._interp_dists[maxidx]
        )
        ln_corr = jnp.where(
            ~mask,
            ln_corr,
            ln_corr[maxidx]
            + slope_at_maxidx * (self._interp_dists - self._interp_dists[maxidx]),
        )

        return Partial(
            self.__class__._interp_cov,
            scale=scale,
            distances=self._interp_dists,
            logcorr=ln_corr,
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022 Max-Planck-Society
# Author: Philipp Arras

import numpy as np

from ..domain_tuple import DomainTuple
from ..sugar import makeField
from ..utilities import myassert
from .linear_operator import LinearOperator


class TransposeOperator(LinearOperator):
    def __init__(self, domain, indices):
        self._domain = DomainTuple.make(domain)
        indices = tuple(indices)
        if len(indices) != len(self._domain):
            raise IndexError("Either too many or too few indices given.")
        self._target = DomainTuple.make(self._domain[ind] for ind in indices)
        if self._domain.size != self._target.size:
            raise ValueError("List of indices not complete")
        self._capability = self._all_ops
        self._np_indices = _niftyspace_to_np_indices(self._domain, indices)
        self._np_indices_inv = tuple(np.argsort(self._np_indices))
        self._indices = indices

    def apply(self, x, mode):
        self._check_input(x, mode)
        x = x.val
        if mode in (self.TIMES, self.ADJOINT_INVERSE_TIMES):
            x = np.transpose(x, self._np_indices)
        else:
            x = np.transpose(x, self._np_indices_inv)
        return makeField(self._tgt(mode), x)

    def __repr__(self):
        return f'Transpose (indices={self._indices})'


def _niftyspace_to_np_indices(domain, indices):
    np_indices = []
    dimensions = np.cumsum((0,) + tuple(len(dd.shape) for dd in domain))
    for ind in indices:
        np_indices.extend(range(dimensions[ind], dimensions[ind+1]))
    res = tuple(np_indices)
    myassert(len(res) == len(domain.shape))
    return res



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .structured_domain import StructuredDomain


class HPSpace(StructuredDomain):
    """Represents 2-sphere with HEALPix discretization.

    Its harmonic partner domain is the
    :class:`~nifty.cl.domains.lm_space.LMSpace`.

    Parameters
    ----------
    nside : int
        The corresponding HEALPix Nside parameter. Must be a positive integer
        and typically is a power of 2.
    """

    _needed_for_hash = ["_nside"]

    def __init__(self, nside):
        self._nside = int(nside)
        if self._nside < 1:
            raise ValueError("nside must be >=1.")

    def __repr__(self):
        return "HPSpace(nside={})".format(self.nside)

    @property
    def harmonic(self):
        return False

    @property
    def shape(self):
        return (self.size,)

    @property
    def size(self):
        return int(12 * self.nside * self.nside)

    @property
    def scalar_dvol(self):
        return np.pi / (3*self._nside*self._nside)

    @property
    def nside(self):
        """int : HEALPix Nside parameter of this domain"""
        return self._nside

    def get_default_codomain(self):
        """Returns a :class:`LMSpace` object, which is capable of storing a
        fairly accurate representation of data residing on `self`

        Returns
        -------
        LMSpace
            The partner domain

        Notes
        -----
        The `lmax` and `mmax` parameters of the returned :class:`LMSpace` are
        set to `2*self.nside`.
        """
        from ..domains.lm_space import LMSpace
        return LMSpace(lmax=2*self.nside)

    def check_codomain(self, codomain):
        """Raises `TypeError` if `codomain` is not a matching partner domain
        for `self`.

        Notes
        -----
        This function only checks whether `codomain` is of type
        :class:`LMSpace`.
        """
        from ..domains.lm_space import LMSpace
        if not isinstance(codomain, LMSpace):
            raise TypeError("codomain must be a LMSpace.")



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import warnings
from functools import partial
from math import log2, sqrt

import jax
import numpy as np
from jax import numpy as jnp
from jax.lax import dynamic_slice_in_dim

from ..tree_math.util import sqrtm
from .util import get_cov_from_loc


def get_1st_hp_nbrs_idx(nside, pix, nest: bool = False, dtype=np.int32):
    from healpy import pixelfunc

    n_nbr = 8

    n_pix = 1 if np.ndim(pix) == 0 else len(pix)
    pix_nbr = np.zeros((n_pix, n_nbr + 1), dtype=int)  # can contain `-1`
    pix_nbr[:, 0] = pix
    nbr = pixelfunc.get_all_neighbours(nside, pix, nest=nest)
    nbr = nbr.reshape(n_nbr, n_pix).T
    pix_nbr[:, 1:] = nbr

    # Account for unknown neighbors, encoded by -1
    idx_w_invalid, nbr_w_invalid = np.nonzero(pix_nbr == -1)
    if idx_w_invalid.size != 0:
        uniq_idx_w_invalid = np.unique(idx_w_invalid)
        nbr_invalid = nbr[uniq_idx_w_invalid]
        with warnings.catch_warnings():
            wmsg = "invalid value encountered in _get_neigbors"
            warnings.filterwarnings("ignore", message=wmsg)
            # shape of (n_2nd_neighbors, n_idx_w_invalid, n_1st_neighbors)
            nbr2 = pixelfunc.get_all_neighbours(nside, nbr_invalid, nest=nest)
        nbr2 = np.transpose(nbr2, (1, 2, 0))
        nbr2[nbr_invalid == -1] = -1
        nbr2 = nbr2.reshape(uniq_idx_w_invalid.size, -1)
        n_replace = np.sum(pix_nbr[uniq_idx_w_invalid] == -1, axis=1)
        if np.any(np.diff(n_replace)):
            raise AssertionError()
        n_replace = n_replace[0]
        pix_2nbr = np.stack(
            [
                np.setdiff1d(ar1, ar2)[:n_replace]
                for ar1, ar2 in zip(nbr2, pix_nbr[uniq_idx_w_invalid])
            ]
        )
        if np.sum(pix_2nbr == -1):
            # `setdiff1d` should remove all `-1` because we worked with rows in
            # pix_nbr that all contain them
            raise AssertionError()
        # Select a "random" 2nd neighbor to fill in for the missing 1st order
        # neighbor
        pix_nbr[idx_w_invalid, nbr_w_invalid] = pix_2nbr.ravel()
        if np.sum(pix_nbr == -1):
            raise AssertionError()

    out = np.squeeze(pix_nbr, axis=0) if np.ndim(pix) == 0 else pix_nbr
    return out.astype(dtype)


def get_all_1st_hp_nbrs_idx(nside, nest: bool = False, dtype=np.int32):
    pix = np.arange(12 * nside**2)
    return get_1st_hp_nbrs_idx(nside, pix, nest=nest, dtype=dtype)


def get_1st_hp_nbrs(nside, pix, nest: bool = False):
    from healpy import pixelfunc

    return np.stack(
        pixelfunc.pix2vec(nside, get_1st_hp_nbrs_idx(nside, pix, nest=nest), nest=nest),
        axis=-1,
    )


def cov_sqrt(chart, kernel, level: int = 0):
    from healpy import pixelfunc

    if chart.ndim != 2:
        nie = "covariance computation only implemented for 3D HEALPix"
        raise NotImplementedError(nie)

    nside = chart.nside_at(level)
    pix0s = np.stack(
        pixelfunc.pix2vec(nside, np.arange(12 * nside**2), nest=chart.nest), axis=-1
    )
    assert pix0s.ndim == 2
    pix0s = pix0s.reshape(pix0s.shape[:1] + (1,) * (chart.ndim - 1) + pix0s.shape[1:])
    if chart.ndim > 1:
        # NOTE, this works for arbitrary many dimensions but is probably not
        # what the user wants. In 1D for example, the distances are still
        # computed for a 3D sphere with unit radius.
        r_rg0 = jnp.mgrid[tuple(slice(s) for s in chart.shape_at(level)[1:])]
        r_rg0 = jnp.asarray(chart.nonhp_ind2cart(r_rg0, level))[..., np.newaxis]
        pix0s = (pix0s * r_rg0).reshape(-1, 3)
    cov_from_loc = get_cov_from_loc(kernel, None)
    return sqrtm(cov_from_loc(pix0s, pix0s))


@partial(jax.jit, static_argnames=("chart", "precision"))
def refine(
    coarse_values,
    excitations,
    olf,
    fks,
    index_map,
    *,
    chart,
    precision=None,
):
    # TODO: Check performance of 'ring' versus 'nest' alignment
    if chart.ndim != 2:
        nie = f"invalid dimensions {chart.ndim!r}; expected `2`"
        raise NotImplementedError(nie)
    coarse_values = coarse_values[:, np.newaxis] if chart.ndim == 1 else coarse_values
    nside = sqrt(coarse_values.shape[0] / 12)
    lvl = int(log2(nside) - log2(chart.nside0))

    def refine(coarse_values, exc, idx_hp, idx_r, olf, fks, im):
        c = dynamic_slice_in_dim(
            coarse_values[chart.hp_neighbors_idx(lvl, idx_hp)],
            idx_r,
            slice_size=chart.coarse_size,
            axis=1,
        )
        f_shp = (chart.fine_size**2, chart.fine_size)
        o = olf[im] if im is not None else olf
        refined = jnp.tensordot(o, c, axes=chart.ndim, precision=precision)
        f = fks[im] if im is not None else fks
        refined += jnp.matmul(f, exc, precision=precision).reshape(f_shp)
        return refined

    pix_hp_idx = jnp.arange(chart.shape_at(lvl)[0])
    pix_r_off = jnp.arange(chart.shape_at(lvl)[1] - chart.coarse_size + 1)
    # TODO: benchmark swapping these two
    off = index_map is not None
    vrefine = jax.vmap(refine, in_axes=(None, 0, None, 0, 0 + off, 0 + off, None))
    in_axes = (None, 0, 0, None)
    in_axes += (0, 0, None) if index_map is None else (None, None, 0)
    vrefine = jax.vmap(vrefine, in_axes=in_axes)
    refined = vrefine(
        coarse_values, excitations, pix_hp_idx, pix_r_off, olf, fks, index_map
    )
    refined = jnp.transpose(refined, (0, 2, 1, 3))
    n_hp = refined.shape[0] * refined.shape[1]
    n_r = refined.shape[2] * refined.shape[3]
    return refined.reshape(n_hp, n_r)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
# Copyright(C) 2025 Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.


import operator
from warnings import warn

import numpy as np
import scipy.fft

from ..config import _config

_nthreads = 1


def nthreads():
    return _nthreads


def set_nthreads(nthr):
    global _nthreads
    _nthreads = int(nthr)


try:
    from pyvkfft.fft import fftn as vkfftn
    from pyvkfft.fft import ifftn as vkifftn
    def _force_complex(a):
        # Vkfft only works on complex inputs
        if a.dtype == np.float32:
            return a.astype(np.complex64)
        if a.dtype == np.float64:
            return a.astype(np.complex128)
        return a
    def cufftn(a, *args, **kwargs):
        return vkfftn(_force_complex(a), *args, **kwargs)
    def cuifftn(a, *args, **kwargs):
        return vkifftn(_force_complex(a), *args, **kwargs)

except ImportError:
    try:
        import cupy

        warn("Using cupy FFT. Consider to install the significantly faster pyvkfft")
        cufftn = cupy.fft.fftn
        cuifftn = cupy.fft.ifftn
    except ImportError:
        pass

def _scipy_fftn(a, axes=None):
    from .any_array import AnyArray
    if a.device_id > -1:
        return AnyArray(cufftn(a._val, axes=axes))
    return AnyArray(scipy.fft.fftn(a._val, axes=axes, workers=_nthreads))


def _scipy_ifftn(a, axes=None):
    from .any_array import AnyArray
    if a.device_id > -1:
        return AnyArray(cuifftn(a._val, axes=axes))
    return AnyArray(scipy.fft.ifftn(a._val, axes=axes, workers=_nthreads))


def _scipy_hartley(a, axes=None):
    from .any_array import AnyArray
    if a.device_id > -1:
        tmp = AnyArray(cufftn(a._val, axes=axes))
    else:
        tmp = AnyArray(scipy.fft.fftn(a._val, axes=axes, workers=_nthreads))
    assert isinstance(tmp, AnyArray)
    c = _config.get("hartley_convention")
    add_or_sub = operator.add if c == "non_canonical_hartley" else operator.sub
    assert isinstance(tmp.real, AnyArray)
    assert isinstance(tmp.imag, AnyArray)
    assert isinstance(add_or_sub(tmp.real, tmp.imag), AnyArray)
    return add_or_sub(tmp.real, tmp.imag)


def _scipy_vdot(a, b):
    from .logger import logger
    if (isinstance(a, np.ndarray) and a.dtype == np.float32) or \
    (isinstance(b, np.ndarray) and b.dtype == np.float32):
        logger.warning("Calling np.vdot in single precision may lead to inaccurate results")
    return np.vdot(a, b)


try:
    import ducc0.fft as my_fft
    import ducc0.misc


    def fftn(a, axes=None):
        from .any_array import AnyArray
        if a.device_id > -1:
            a = cufftn(a._val, axes=axes)
        else:
            a = my_fft.c2c(a._val, axes=axes, nthreads=max(_nthreads, 0))
        return AnyArray(a)


    def ifftn(a, axes=None):
        from .any_array import AnyArray
        if a.device_id > -1:
            a = cuifftn(a._val, axes=axes)
        else:
            a = my_fft.c2c(a._val, axes=axes, inorm=2, forward=False,
                           nthreads=max(_nthreads, 0))
        return AnyArray(a)


    def hartley(a, axes=None):
        from .any_array import AnyArray
        if a.device_id > -1:
            return _scipy_hartley(a, axes)
        c = _config.get("hartley_convention")
        ht = my_fft.genuine_hartley if c == "non_canonical_hartley" else my_fft.genuine_fht
        a = ht(a._val, axes=axes, nthreads=max(_nthreads, 0))
        return AnyArray(a)


    def vdot(a, b):
        if isinstance(a, np.ndarray) and np.issubdtype(a.dtype, np.integer):
            a = a.astype(np.float64)
        if isinstance(b, np.ndarray) and np.issubdtype(b.dtype, np.integer):
            b = b.astype(np.float64)
        return ducc0.misc.vdot(a, b)

except ImportError:
    fftn = _scipy_fftn
    ifftn = _scipy_ifftn
    hartley = _scipy_hartley
    vdot = _scipy_vdot



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .domain import Domain


class StructuredDomain(Domain):
    """The abstract base class for all structured domains.

    An instance of a space contains information about the manifold's
    geometry and enhances the functionality of Domain by methods that
    are needed for power spectrum analysis and smoothing.
    """

    @property
    def scalar_dvol(self):
        """float or None : uniform pixel volume, if applicable

        Returns the volume factors of this domain as a floating
        point scalar, if the volume factors are all identical, otherwise
        returns None.
        """
        raise NotImplementedError

    @property
    def dvol(self):
        """float or numpy.ndarray(dtype=float): pixel volume(s)

        Returns the volume factors of this domain, either as a floating
        point scalar (if the volume factors are all identical) or as a
        floating point array with a shape of `self.shape`.
        """
        return self.scalar_dvol

    @property
    def total_volume(self):
        """float : Total domain volume.

        Returns the sum over all the domain's pixel volumes.
        """
        tmp = self.dvol
        return self.size * tmp if np.isscalar(tmp) else np.sum(tmp)

    @property
    def harmonic(self):
        """bool : True iff this domain is a harmonic domain."""
        raise NotImplementedError

    def get_k_length_array(self):
        """k vector lengths, if applicable.

        Returns the length of the k vector for every pixel.
        This method is only implemented for harmonic domains.

        Returns
        -------
        :class:`nifty.cl.field.Field`
            An array containing the k vector lengths
        """
        raise NotImplementedError

    def get_unique_k_lengths(self):
        """Sorted unique k-vector lengths, if applicable.

        Returns an array of floats containing the unique k vector lengths
        for this domain.
        This method is only implemented for harmonic domains.
        """
        raise NotImplementedError

    def get_fft_smoothing_kernel_function(self, sigma):
        """Helper for Gaussian smoothing.

        This method, which is only implemented for harmonic domains, helps to
        smoothe fields that are defined on a domain that has this domain as
        its harmonic partner. The returned function does a pointwise evaluation
        of a zero-centered Gaussian on the field values, which corresponds to a
        convolution with a Gaussian kernel with sigma standard deviation in
        position space.

        Parameters
        ----------
        sigma : float
            A real number representing a physical scale on which the smoothing
            takes place. Mathematically sigma is the standard
            deviation of a convolution with a normalized, zero-centered
            Gaussian that takes place in position space.

        Returns
        -------
        function (array-like -> array-like)
            A smoothing operation that multiplies values with a Gaussian
            kernel.
        """
        raise NotImplementedError



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import warnings
from functools import wraps
from typing import Any, Callable, Dict, Hashable, Mapping, ParamSpec, TypeVar

import jax
import numpy as np
from jax import numpy as jnp
from jax.tree_util import Partial

O = TypeVar("O")
I = TypeVar("I")


def hvp(f, primals, tangents):
    return jax.jvp(jax.grad(f), primals, tangents)[1]


def split(mappable, keys):
    """Split a dictionary into one containing only the specified keys and one
    with all of the remaining ones.
    """
    sel, rest = {}, {}
    for k, v in mappable.items():
        if k in keys:
            sel[k] = v
        else:
            rest[k] = v
    return sel, rest


def isiterable(candidate):
    try:
        iter(candidate)
        return True
    except (TypeError, AttributeError):
        return False


def is_iterable_of_non_iterables(ls: Any) -> bool:
    """Indicates whether the input is one dimensional.

    An object is considered one dimensional if it is an iterable of
    non-iterable items.
    """
    if hasattr(ls, "ndim"):
        return ls.ndim == 1
    if not isiterable(ls):
        return False
    return all(not isiterable(e) for e in ls)


def doc_from(original):
    def wrapper(target):
        target.__doc__ = original.__doc__
        return target

    return wrapper


def wrap(
    call: Callable[[I], O],
    name: Hashable,
) -> Callable[[Mapping[Hashable, I]], O]:
    def named_call(p):
        return call(p[name])

    return named_call


def wrap_left(
    call: Callable[[I], O],
    name: Hashable,
) -> Callable[[I], Dict[Hashable, O]]:
    def named_call(p):
        return {name: call(p)}

    return named_call


def interpolate(xmin=-7.0, xmax=7.0, N=14000) -> Callable:
    """Replaces a local nonlinearity such as jnp.exp with a linear interpolation

    Interpolating functions speeds up code and increases numerical stability in
    some cases, but at a cost of precision and range.

    Parameters
    ----------
    xmin : float
        Minimal interpolation value. Default: -7.
    xmax : float
        Maximal interpolation value. Default: 7.
    N : int
        Number of points used for the interpolation. Default: 14000
    """

    def decorator(f):
        x = jnp.linspace(xmin, xmax, N)
        y = f(x)

        @wraps(f)
        def wrapper(t):
            return jnp.interp(t, x, y)

        return wrapper

    return decorator


def _to_numpy(arg):
    # NOTE, assume no cycles in the input and recurse into it without safeguards
    if isinstance(arg, jax.Array):
        return np.asarray(arg)
    elif isinstance(arg, dict):
        # JAX arrays are not hashable so keys do not need to be checked
        return {k: _to_numpy(v) for k, v in arg.items()}
    elif isinstance(arg, (tuple, list)):
        type(arg)(map(_to_numpy, arg))
    return arg


def safeguard_arguments_against_accidental_calls_into_jax(func):
    """Safeguard against using JAX in callback, see !25861"""

    @wraps(func)
    def safe_func(*args, **kwargs):
        return func(*_to_numpy(args), **_to_numpy(kwargs))

    return safe_func


@safeguard_arguments_against_accidental_calls_into_jax
def _maybe_raise(condition, exception):
    if condition:
        raise exception()


def conditional_raise(condition: bool, exception):
    """JAX JIT-safe raise of the given Exception if `condition` is True.

    Parameters:
    -----------
    condition: bool
        If True, will raise `exception` on the host.
    exception: :class:`Exception`
        Exception that will be raised if `condition` is True
    """
    from jax.debug import callback

    from .tree_math import hide_strings

    # Register as few host-callbacks as possible by implicitly hashing the
    # exception type and the strings within
    callback(
        _maybe_raise,
        condition,
        Partial(exception.__class__, *hide_strings(exception.args)),
    )


@safeguard_arguments_against_accidental_calls_into_jax
def _maybe_call(condition, fn, args, kwargs):
    if condition:
        fn(*args, **kwargs)


def conditional_call(condition, fn, *args, **kwargs):
    """JAX JIT-safe call to `fn` if `condition` is True.

    Warning:
    ---------
    The function `fn` may NOT dispatch ANY JAX code, including by comparing JAX
    objects! To safeguard against easy to miss fallacies, all JAX arrays are
    automatically converted to numpy arrays.

    Parameters:
    -----------
    condition: boolean
        If True, will call `fn` on the host
    fn: Callable
        Function that will be called on the host if `condition` is True
    args:
        Positional arguments passed to `fn`
    kwargs:
        Keyword arguments passed to `fn`
    """
    from jax.debug import callback

    callback(_maybe_call, condition, Partial(fn), args, kwargs)


_ReturnType = TypeVar("rT")  # return type
_ParameterType = ParamSpec("pT")  # parameters type


def deprecated(message: str):
    def _deprecate(
        func: Callable[_ParameterType, _ReturnType],
    ) -> Callable[_ParameterType, _ReturnType]:
        @wraps(func)
        def wrapped_func(*args: _ParameterType.args, **kwargs: _ParameterType.kwargs):
            warnings.simplefilter("always", DeprecationWarning)  # turn off filter
            warnings.warn(message, category=DeprecationWarning, stacklevel=2)
            warnings.simplefilter("default", DeprecationWarning)  # reset filter
            return func(*args, **kwargs)

        return wrapped_func

    return _deprecate



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import functools
from time import time

import numpy as np

from ..logger import logger
from ..utilities import NiftyMeta


class IterationController(metaclass=NiftyMeta):
    """The abstract base class for all iteration controllers.
    An iteration controller is an object that monitors the progress of a
    minimization iteration. At the begin of the minimization, its start()
    method is called with the energy object at the initial position.
    Afterwards, its check() method is called during every iteration step with
    the energy object describing the current position.
    Based on that information, the iteration controller has to decide whether
    iteration needs to progress further (in this case it returns CONTINUE), or
    if sufficient convergence has been reached (in this case it returns
    CONVERGED), or if some error has been detected (then it returns ERROR).

    The concrete convergence criteria can be chosen by inheriting from this
    class; the implementer has full flexibility to use whichever criteria are
    appropriate for a particular problem - as long as they can be computed from
    the information passed to the controller during the iteration process.

    For analyzing minimization procedures IterationControllers can log energy
    values together with the respective time stamps. In order to activate this
    feature `enable_logging()` needs to be called.
    """

    CONVERGED, CONTINUE, ERROR = list(range(3))

    def __init__(self):
        self._history = None

    def start(self, energy):
        """Starts the iteration.

        Parameters
        ----------
        energy : Energy object
           Energy object at the start of the iteration

        Returns
        -------
        status : integer status, can be CONVERGED, CONTINUE or ERROR
        """
        raise NotImplementedError

    def check(self, energy):
        """Checks the state of the iteration. Called after every step.

        Parameters
        ----------
        energy : Energy object
           Energy object at the start of the iteration

        Returns
        -------
        status : integer status, can be CONVERGED, CONTINUE or ERROR
        """
        raise NotImplementedError

    def enable_logging(self):
        """Enables the logging functionality. If the log has been populated
        before, it stays as it is."""
        if self._history is None:
            self._history = EnergyHistory()

    def disable_logging(self):
        """Disables the logging functionality. If the log has been populated
        before, it is dropped."""
        self._history = None

    @property
    def history(self):
        return self._history


class EnergyHistory:
    def __init__(self):
        self._lst = []

    def append(self, x):
        if len(x) != 2:
            raise ValueError
        self._lst.append((float(x[0]), float(x[1])))

    def reset(self):
        self._lst = []

    def __getitem__(self, i):
        return self._lst[i]

    @property
    def time_stamps(self):
        return [x for x, _ in self._lst]

    @property
    def energy_values(self):
        return [x for _, x in self._lst]

    def __add__(self, other):
        if not isinstance(other, EnergyHistory):
            return NotImplemented
        res = EnergyHistory()
        res._lst = self._lst + other._lst
        return res

    def __iadd__(self, other):
        if not isinstance(other, EnergyHistory):
            return NotImplemented
        self._lst += other._lst
        return self

    def __len__(self):
        return len(self._lst)


def append_history(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        hist = args[0].history
        if isinstance(hist, EnergyHistory):
            hist.append((time(), args[1].value))
        return func(*args, **kwargs)
    return wrapper


class GradientNormController(IterationController):
    """An iteration controller checking (mainly) the L2 gradient norm.

    Parameters
    ----------
    tol_abs_gradnorm : float, optional
        If the L2 norm of the energy gradient is below this value, the
        convergence counter will be increased in this iteration.
    tol_rel_gradnorm : float, optional
        If the L2 norm of the energy gradient divided by its initial L2 norm
        is below this value, the convergence counter will be increased in this
        iteration.
    convergence_level : int, default=1
        The number which the convergence counter must reach before the
        iteration is considered to be converged
    iteration_limit : int, optional
        The maximum number of iterations that will be carried out.
    name : str, optional
        if supplied, this string and some diagnostic information will be
        printed after every iteration
    """

    def __init__(self, tol_abs_gradnorm=None, tol_rel_gradnorm=None,
                 convergence_level=1, iteration_limit=None, name=None):
        super(GradientNormController, self).__init__()
        self._tol_abs_gradnorm = tol_abs_gradnorm
        self._tol_rel_gradnorm = tol_rel_gradnorm
        self._convergence_level = convergence_level
        self._iteration_limit = iteration_limit
        self._name = name

    @append_history
    def start(self, energy):
        self._itcount = -1
        self._ccount = 0
        if self._tol_rel_gradnorm is not None:
            self._tol_rel_gradnorm_now = self._tol_rel_gradnorm \
                                       * energy.gradient_norm
        return self.check(energy)

    @append_history
    def check(self, energy):
        self._itcount += 1

        inclvl = False
        if self._tol_abs_gradnorm is not None:
            if energy.gradient_norm <= self._tol_abs_gradnorm:
                inclvl = True
        if self._tol_rel_gradnorm is not None:
            if energy.gradient_norm <= self._tol_rel_gradnorm_now:
                inclvl = True
        if inclvl:
            self._ccount += 1
        else:
            self._ccount = max(0, self._ccount-1)

        # report
        if self._name is not None:
            logger.info(
                "{}: Iteration #{} energy={:.6E} gradnorm={:.2E} clvl={}"
                .format(self._name, self._itcount, energy.value,
                        energy.gradient_norm, self._ccount))

        # Are we done?
        if self._iteration_limit is not None:
            if self._itcount >= self._iteration_limit:
                logger.warning(
                    "{}Iteration limit reached. Assuming convergence"
                    .format("" if self._name is None else self._name+": "))
                return self.CONVERGED
        if self._ccount >= self._convergence_level:
            return self.CONVERGED

        return self.CONTINUE


class GradInfNormController(IterationController):
    """An iteration controller checking (mainly) the L_infinity gradient norm.

    Parameters
    ----------
    tol : float
        If the L_infinity norm of the energy gradient is below this value, the
        convergence counter will be increased in this iteration.
    convergence_level : int, default=1
        The number which the convergence counter must reach before the
        iteration is considered to be converged
    iteration_limit : int, optional
        The maximum number of iterations that will be carried out.
    name : str, optional
        if supplied, this string and some diagnostic information will be
        printed after every iteration
    """

    def __init__(self, tol, convergence_level=1, iteration_limit=None,
                 name=None):
        super(GradInfNormController, self).__init__()
        self._tol = tol
        self._convergence_level = convergence_level
        self._iteration_limit = iteration_limit
        self._name = name

    @append_history
    def start(self, energy):
        self._itcount = -1
        self._ccount = 0
        return self.check(energy)

    @append_history
    def check(self, energy):
        self._itcount += 1

        crit = energy.gradient.norm(np.inf) / abs(energy.value)
        if self._tol is not None and crit <= self._tol:
            self._ccount += 1
        else:
            self._ccount = max(0, self._ccount-1)

        # report
        if self._name is not None:
            logger.info(
                "{}: Iteration #{} energy={:.6E} crit={:.2E} clvl={}"
                .format(self._name, self._itcount, energy.value,
                        crit, self._ccount))

        # Are we done?
        if self._iteration_limit is not None:
            if self._itcount >= self._iteration_limit:
                logger.warning(
                    "{} Iteration limit reached. Assuming convergence"
                    .format("" if self._name is None else self._name+": "))
                return self.CONVERGED
        if self._ccount >= self._convergence_level:
            return self.CONVERGED

        return self.CONTINUE


class DeltaEnergyController(IterationController):
    """An iteration controller checking (mainly) the relative energy change
    from one iteration to the next.

    Parameters
    ----------
    tol_rel_deltaE : float
        If the difference between the last and current energies divided by
        the current energy is below this value, the convergence counter will
        be increased in this iteration.
    convergence_level : int, default=1
        The number which the convergence counter must reach before the
        iteration is considered to be converged
    iteration_limit : int, optional
        The maximum number of iterations that will be carried out.
    name : str, optional
        if supplied, this string and some diagnostic information will be
        printed after every iteration
    """

    def __init__(self, tol_rel_deltaE, convergence_level=1,
                 iteration_limit=None, name=None):
        super(DeltaEnergyController, self).__init__()
        self._tol_rel_deltaE = tol_rel_deltaE
        self._convergence_level = convergence_level
        self._iteration_limit = iteration_limit
        self._name = name

    @append_history
    def start(self, energy):
        self._itcount = -1
        self._ccount = 0
        self._Eold = 0.
        return self.check(energy)

    @append_history
    def check(self, energy):
        self._itcount += 1

        inclvl = False
        Eval = energy.value
        rel = abs(self._Eold-Eval)/max(abs(self._Eold), abs(Eval))
        if self._itcount > 0:
            if rel < self._tol_rel_deltaE:
                inclvl = True
        self._Eold = Eval
        if inclvl:
            self._ccount += 1
        else:
            self._ccount = max(0, self._ccount-1)

        # report
        if self._name is not None:
            logger.info(
                "{}: Iteration #{} energy={:.6E} reldiff={:.6E} clvl={}"
                .format(self._name, self._itcount, Eval, rel, self._ccount))

        # Are we done?
        if self._iteration_limit is not None:
            if self._itcount >= self._iteration_limit:
                logger.warning(
                    "{} Iteration limit reached. Assuming convergence"
                    .format("" if self._name is None else self._name+": "))
                return self.CONVERGED
        if self._ccount >= self._convergence_level:
            return self.CONVERGED

        return self.CONTINUE


class AbsDeltaEnergyController(IterationController):
    """An iteration controller checking (mainly) the energy change from one
    iteration to the next.

    Parameters
    ----------
    deltaE : float
        If the difference between the last and current energies is below this
        value, the convergence counter will be increased in this iteration.
    convergence_level : int, default=1
        The number which the convergence counter must reach before the
        iteration is considered to be converged
    iteration_limit : int, optional
        The maximum number of iterations that will be carried out.
    name : str, optional
        if supplied, this string and some diagnostic information will be
        printed after every iteration
    """

    def __init__(self, deltaE, convergence_level=1, iteration_limit=None,
                 name=None):
        super(AbsDeltaEnergyController, self).__init__()
        self._deltaE = deltaE
        self._convergence_level = convergence_level
        self._iteration_limit = iteration_limit
        self._name = name

    @append_history
    def start(self, energy):
        self._itcount = -1
        self._ccount = 0
        self._Eold = 0.
        return self.check(energy)

    @append_history
    def check(self, energy):
        self._itcount += 1

        inclvl = False
        Eval = energy.value
        diff = abs(self._Eold-Eval)
        if self._itcount > 0:
            if diff < self._deltaE:
                inclvl = True
        self._Eold = Eval
        if inclvl:
            self._ccount += 1
        else:
            self._ccount = max(0, self._ccount-1)

        # report
        if self._name is not None:
            logger.info(
                "{}: Iteration #{} energy={:.6E} diff={:.6E} crit={:.1E} clvl={}"
                .format(self._name, self._itcount, Eval, diff, self._deltaE,
                        self._ccount))

        # Are we done?
        if self._iteration_limit is not None:
            if self._itcount >= self._iteration_limit:
                logger.warning(
                    "{} Iteration limit reached. Assuming convergence"
                    .format("" if self._name is None else self._name+": "))
                return self.CONVERGED
        if self._ccount >= self._convergence_level:
            return self.CONVERGED

        return self.CONTINUE


class StochasticAbsDeltaEnergyController(IterationController):
    """Check the standard deviation over a period of iterations.

    Convergence is reported once this quantity falls below the given threshold.


    Parameters
    ----------
    deltaE : float
        If the standard deviation of the last energies is below this
        value, the convergence counter will be increased in this iteration.
    convergence_level : int, optional
        The number which the convergence counter must reach before the
        iteration is considered to be converged. Defaults to 1.
    iteration_limit : int, optional
        The maximum number of iterations that will be carried out.
    name : str, optional
        If supplied, this string and some diagnostic information will be
        printed after every iteration.
    memory_length : int, optional
        The number of last energies considered for determining convergence,
        defaults to 10.
    """

    def __init__(self, deltaE, convergence_level=1, iteration_limit=None,
                 name=None, memory_length=10):
        super(StochasticAbsDeltaEnergyController, self).__init__()
        self._deltaE = deltaE
        self._convergence_level = convergence_level
        self._iteration_limit = iteration_limit
        self._name = name
        self.memory_length = memory_length

    @append_history
    def start(self, energy):
        self._itcount = -1
        self._ccount = 0
        self._memory = []
        return self.check(energy)

    @append_history
    def check(self, energy):
        self._itcount += 1

        inclvl = False
        Eval = energy.value
        self._memory.append(Eval)
        if len(self._memory) > self.memory_length:
            self._memory = self._memory[1:]
        diff = np.std(self._memory)
        if self._itcount > 0:
            if diff < self._deltaE:
                inclvl = True
        if inclvl:
            self._ccount += 1
        else:
            self._ccount = max(0, self._ccount-1)

        # report
        if self._name is not None:
            logger.info(
                "{}: Iteration #{} energy={:.6E} diff={:.6E} crit={:.1E} clvl={}"
                .format(self._name, self._itcount, Eval, diff, self._deltaE,
                        self._ccount))

        # Are we done?
        if self._iteration_limit is not None:
            if self._itcount >= self._iteration_limit:
                logger.warning(
                    "{} Iteration limit reached. Assuming convergence"
                    .format("" if self._name is None else self._name+": "))
                return self.CONVERGED
        if self._ccount >= self._convergence_level:
            return self.CONVERGED

        return self.CONTINUE



#!/usr/bin/env python
# coding: utf-8
# # Showcasing the Correlated Field Model
#
# The field model works roughly like this:
#
# ``f = HT( A * zero_mode * xi ) + offset``
#
# The correlated field is constructed using:
#
# \begin{equation*}
# cf = \verb|offset_mean| + \left(\bigotimes_{i} \frac{1}{V_i} HT_i \right) \left( \verb|zero_mode| \cdot (\bigotimes_{i} A_i (k))
# \right)
# \end{equation*}
#
# where the outer product $\bigotimes_{i}$ is taken over all subdomains, $V_i$ is the volume of each sub-space, $HT_i$ is the harmonic transform over each sub-space
#
# `A`  is a spectral power field which is constructed from power spectra that are defined on subdomains of the target domain. It is scaled by a zero mode operator and then pointwise multiplied by a Gaussian excitation field, yielding a representation of the field in harmonic space. It is then transformed into the target real space and a offset is added.
#
# The power spectra that `A` is constructed of, are in turn constructed as the sum of a power law component and an integrated Wiener process whose amplitude and roughness can be set.

# ## Preliminaries

# +
# Imports and Initializing dimensions and a seed

# %matplotlib inline
import jax
import nifty.re as jft
import matplotlib.pyplot as plt
from jax import numpy as jnp
from typing import Tuple
import numpy as np

jax.config.update("jax_enable_x64", True)

plt.rcParams["figure.dpi"] = 300

npix = 256
seed = 42
distances = 1
key = jax.random.PRNGKey(seed)
k_lengths = jnp.arange(0, npix) * (1 / npix)
totvol = jnp.prod(jnp.array(npix) * jnp.array(distances))
realisations = 5


# ## The Moment Matched Log-Normal Distribution
#
# Many properties of the correlated field are modelled as being lognormally distributed.
#
# The distribution models are parametrized via their means and standard-deviations (first and second position in tuple).
#
# To get a feeling of how the ratio of the `mean` and `stddev` parameters influences the distribution shape, here are a few example histograms: (observe the x-axis!)

# +
fig = plt.figure(figsize=(13, 3.5))
mean = 1.0
sigmas = [1.0, 0.5, 0.1]


for i in range(3):
    op = jft.LogNormalPrior(mean=mean, std=sigmas[i], name="foo")
    op_samples = np.zeros(10000)
    for j in range(10000):
        key, signalk = jax.random.split(key, num=2)
        s = jft.random_like(signalk, op.domain)
        op_samples[j] = op(s)

    ax = fig.add_subplot(1, 3, i + 1)
    ax.hist(op_samples, bins=50)
    ax.set_title(f"mean = {mean}, sigma = {sigmas[i]}")
    ax.set_xlabel("x")
    del op_samples


plt.show()

# +
# Making the Correlated Field in Nifty.re


def fieldmaker(npix: Tuple, distances: Tuple, matern, **args):
    cf = jft.CorrelatedFieldMaker("")
    cf.set_amplitude_total_offset(
        offset_mean=args["offset_mean"], offset_std=args["offset_std"]
    )
    args.pop("offset_mean")
    args.pop("offset_std")
    # There are two choices to the kwarg non_parametric_kind, power and amplitude. NIFTy.re's default is amplitude.
    cf.add_fluctuations_matern(
        npix,
        distances,
        non_parametric_kind="power",
        renormalize_amplitude=False,
        **args,
    ) if matern else cf.add_fluctuations(
        npix, distances, non_parametric_kind="power", **args
    )
    cf_model = cf.finalize()

    return cf_model, cf.power_spectrum


def vary_parameter(parameter, values, matern, **args):
    global key
    for i, j in enumerate(values):
        syn_data = np.zeros(shape=(npix, realisations))
        syn_pow = np.zeros(shape=(int(npix / 2 + 1), realisations))
        args[parameter] = j
        fig = plt.figure(tight_layout=True, figsize=(10, 3))
        fig.suptitle(f"{parameter} = {j}")
        ax1 = fig.add_subplot(1, 2, 1)
        ax1.set_title("Field Realizations")
        ax1.set_ylim(
            -4.0,
            4,
        )
        ax2 = fig.add_subplot(1, 2, 2)
        ax2.set_xscale("log")
        ax2.set_yscale("log")
        ax2.set_title("Power Spectra")
        # Plotting different realisations for each field.
        for k in range(realisations):
            cf_model, pow_cf = fieldmaker(npix, distances, matern, **args)
            key, signalk = jax.random.split(key, num=2)
            syn_signal = jft.random_like(signalk, cf_model.domain)
            syn_data[:, k] = cf_model(syn_signal)
            syn_pow[:, k] = pow_cf(syn_signal)
            ax1.plot(k_lengths, syn_data[:, k], linewidth=1)
            if not matern:
                ax2.plot(
                    np.arange(len(k_lengths) / 2 + 1),
                    np.sqrt(syn_pow[:, k] / totvol ** 2),
                    linewidth=1,
                )
                ax2.set_ylim(1e-6, 2.0)
            else:
                ax2.plot(np.arange(len(k_lengths) / 2 + 1), syn_pow[:, k], linewidth=1)
                ax2.set_ylim(1e-1, 1e2)


# -

# ### The Amplitude Spectrum in NIFTy.re
#
# The correlation operator $A$ which used to transform a to-be-inferred signal $s$ to standardized $\xi$ coordinates, is given by
# $ s = A \xi $
#
# and A is defined as (see below for derivation)
#
# $$ A \mathrel{\mathop{:}}= F^{-1}\hat{S}^{1/2} $$
#
# Where $F^{-1}$ is the inverse Fourier transform, and $\hat{S}$ is the diagonalized correlation structure in harmonic space.
#
# The norm is defined as:
#
# \begin{equation*}
#     \text{norm} = \sqrt{\frac{\int \textnormal{d}k p(k)}{V}}
# \end{equation*}
#
# The amplitude spectrum $\text{amp(k)}$ is then
# \begin{align*}
#     \text{amp(k)} = \frac{\text{fluc} \cdot \sqrt{V}}{\text{norm}} \cdot \sqrt{p(k)} \\
#     \text{amp(k)} = \frac{\text{fluc} \cdot V \cdot \sqrt{p(k)}}{\sqrt{\int \textnormal{d}k p(k)}}
# \end{align*}
#
# The power spectrum is just the amplitude spectrum squared:
#
# \begin{align*}
#     p(k) &= \text{amp}^2(k) \\
#     & = \frac{\text{fluc}^2 \cdot V^2 \cdot p(k)}{\int \textnormal{d}k p(k)} \\
#     \int \textnormal{d}k p(k) & = \text{fluc}^2 \cdot V^2
# \end{align*}
#
# Hence, the fluctuations in `NIFTy.re` are given by
#
# \begin{equation*}
#     \text{fluc} = \sqrt{\frac{\int \textnormal{d}k p(k)}{V^2}}
# \end{equation*}
#
# This is different as compared to the Numpy-based NIFTy, where:
#
# \begin{equation*}
#     \text{fluc} = \int \textnormal{d}k p(k)
# \end{equation*}
#
#

# ## The Neutral Field

# +
cf_args = {
    "fluctuations": (1e-3, 1e-16),
    "loglogavgslope": (0.0, 1e-16),
    "flexibility": (1e-3, 1e-16),
    "asperity": (1e-3, 1e-16),
    "prefix": "",
    "offset_mean": 0.0,
    "offset_std": (1e-3, 1e-16),
}

cf_model, pow_cf = fieldmaker(npix, distances, matern=False, **cf_args)

key, signalk = jax.random.split(key, num=2)
syn_signal = jft.random_like(signalk, cf_model.domain)
syn_data = cf_model(syn_signal)
syn_pow = pow_cf(syn_signal)
fig = plt.figure(tight_layout=True, figsize=(10, 3))
ax1 = fig.add_subplot(1, 2, 1)
ax1.set_title("Field Realizations")
ax1.set_ylim(
    -4.0,
    4,
)
ax2 = fig.add_subplot(1, 2, 2)
ax2.set_xscale("log")
ax2.set_yscale("log")
ax2.set_title("Power Spectra")
ax2.set_ylim(1e-6, 2.0)
ax1.plot(k_lengths, syn_data, linewidth=1)
ax2.plot(np.arange(len(k_lengths) / 2 + 1), np.sqrt(syn_pow / totvol ** 2), linewidth=1)

# -

# ## The `fluctuations` parameters of `add_fluctuations()`
#
# `fluctuations` determine the **amplitude of variations** along the field dimension for which add_fluctuations is called.
#
# `fluctuations[0]` set the average amplitude of the fields fluctuations along the given dimension,
# `fluctuations[1]` sets the width and shape of the amplitude distribution.
#

# ## `fluctuations` mean

vary_parameter(
    "fluctuations", [(0.05, 1e-16), (0.5, 1e-16), (1.0, 1e-16)], matern=False, **cf_args
)
cf_args["fluctuations"] = (1.0, 1e-16)

# ## `fluctuations` std

vary_parameter(
    "fluctuations", [(1.0, 0.01), (1.0, 0.1), (1.0, 1.0)], matern=False, **cf_args
)
cf_args["fluctuations"] = (1.0, 1e-16)


# ## The `loglogavgslope` parameters of `add_fluctuations()`
#
# The value of `loglogavgslope` determines the __slope of the loglog-linear (power law) component of the power spectrum.__
#
# The slope is modelled to be normally distributed.

# ## `loglogavgslope` mean

vary_parameter(
    "loglogavgslope",
    [(-6.0, 1e-16), (-2.0, 1e-16), (2.0, 1e-16)],
    matern=False,
    **cf_args,
)

# ## `loglogavgslope` std

vary_parameter(
    "loglogavgslope", [(-2.0, 0.02), (-2.0, 0.2), (-2.0, 2.0)], matern=False, **cf_args
)
cf_args["loglogavgslope"] = (-2.0, 1e-16)

# ## The `flexibility` parameters of `add_fluctuations()`
#
# Values for `flexibility` determine the __amplitude of the integrated Wiener process component of the power spectrum__ (how strong the power spectrum varies besides the power-law).
#
# `flexibility[0]` sets the _average_ amplitude of the i.g.p. component,
# `flexibility[1]` sets how much the amplitude can vary.
# These two parameters feed into a moment-matched log-normal distribution model, see above for a demo of its behavior.

# ## `flexibility` mean

vary_parameter(
    "flexibility", [(0.4, 1e-16), (4.0, 1e-16), (12.0, 1e-16)], matern=False, **cf_args
)

# ## `flexibility` std

vary_parameter(
    "flexibility", [(4.0, 0.02), (4.0, 0.2), (4.0, 2.0)], matern=False, **cf_args
)
cf_args["flexibility"] = (4.0, 1e-16)

# ## The `asperity` parameters of `add_fluctuations()`
#
# `asperity` determines how __rough the integrated Wiener process component of the power spectrum is.__
#
# `asperity[0]` sets the average roughness, `asperity[1]` sets how much the roughness can vary.
# These two parameters feed into a moment-matched log-normal distribution model, see above for a demo of its behavior.
#
#

# ## `asperity` mean

vary_parameter(
    "asperity", [(0.001, 1e-16), (1.0, 1e-16), (5.0, 1e-16)], matern=False, **cf_args
)

# ## `asperity` std

vary_parameter(
    "asperity", [(1.0, 0.01), (1.0, 0.1), (1.0, 1.0)], matern=False, **cf_args
)
cf_args["asperity"] = (1.0, 1e-16)

# ## The `offset_mean` parameter of `CorrelatedFieldMaker()`
#
# The `offset_mean` parameter defines a global additive offset on the field realizations.
#
# If the field is used for a lognormal model `f = field.exp()`, this acts as a global signal magnitude offset.
#
#

# Reset model to neutral
cf_args["fluctuations"] = (1e-3, 1e-16)
cf_args["flexibility"] = (1e-3, 1e-16)
cf_args["asperity"] = (1e-3, 1e-16)
cf_args["loglogavgslope"] = (1e-3, 1e-16)

vary_parameter("offset_mean", [3.0, 0.0, -2.0], matern=False, **cf_args)

vary_parameter(
    "offset_std", [(1e-16, 1e-16), (0.5, 1e-16), (2.0, 1e-16)], matern=False, **cf_args
)

vary_parameter(
    "offset_std", [(1.0, 0.01), (1.0, 0.1), (1.0, 1.0)], matern=False, **cf_args
)

# ## Matern Fluctuation Kernels
#
# The correlated fields model also supports parametrizing the power spectra of field dimensions using Matern kernels. In the following, the effects of their parameters are demonstrated.
#
# Contrary to the field fluctuations parametrization showed above, the Matern kernel parameters show strong interactions. For example, the field amplitude does not only depend on the amplitude scaling parameter `scale`, but on the combination of all three parameters `scale`, `cutoff` and `loglogavgslope`.

# +
# Neutral model parameters yielding a quasi-constant field

cf_args_matern = {
    "scale": (1e-2, 1e-16),
    "cutoff": (1.0, 1e-16),
    "loglogslope": (-2.0, 1e-16),
    "prefix": "",
    "offset_mean": 0.0,
    "offset_std": (1e-3, 1e-16),
}
# -

vary_parameter(
    "scale", [(0.01, 1e-16), (0.1, 1e-16), (1.0, 1e-16)], matern=True, **cf_args_matern
)

vary_parameter(
    "scale", [(0.5, 0.01), (0.5, 0.1), (0.5, 0.5)], matern=True, **cf_args_matern
)
cf_args_matern["scale"] = (0.5, 1e-16)

vary_parameter(
    "cutoff", [(10.0, 1.0), (10.0, 3.16), (10.0, 100.0)], matern=True, **cf_args_matern
)

# ## Theory of the Correlated Field Model
#
# - Want to infer a signal $s$, and have a (multivariate Gaussian) prior $\mathcal{G}(s,S)$
# - Inference algorithms are sensitive to coordinates, more efficient to infer $\mathcal{G}(\xi, 1)$ as opposed to $\mathcal{G}(s,S)$
# - Can find a coordinate transform (known as the amplitude transform) which changes the distribution to standardized coordinates $\xi \hookleftarrow \mathcal{G}(\xi, 1)$.
# - This transform is given by $ s = F^{-1}\hat{S}^{1/2}\xi \mathrel{\mathop{:}}= A\xi $
#
# We can now define a Correlation Operator $A$ as
# \begin{equation}
#     A \mathrel{\mathop{:}}= F^{-1}\hat{S}^{1/2}
# \end{equation}
# which relates $\xi$ from the latent space to $s$ in physical space.
#
# The correlation structure $S$ and $A$ are related as,
# \begin{equation}
#     S = A A^{\dagger}
# \end{equation}
#
#
# (Here we assume that the correlation structure $S$ is statistically homogenous and stationary and isotropic (which means that $S$ does not have a preferred location or direction a priori), which then according to the Wiener-Khintchin theorem can be represented as a diagonal matrix in Harmonic (Fourier) Space.)
#
# Then,
#
# $S$ can be written as $S = F^\dagger (\hat{p_S}) F $ where $\hat{p}$ is a diagonal matrix with the power spectra values for each mode on the diagonal.
#
# From the relation given above, the amplitude spectrum $p_A$ is related to the power spectrum as:
# $$ p_A(k) = \sqrt{p_S(k)} $$
# $$ C_s(k) = \lim\limits_{V \to \infty} \frac{1}{V} \left< \left| \int_{V} dx s^{x} e^{ikx} \right| \right>_{s}$$
#
#
# If we do not have enough information about the correlation structure $S$ (and consequently the amplitude $A$), we build a model for it and learn it using data. Again, assuming statistical homogeneity a priori, then the Wiener Khintchin theorem yields:
#
# $$ A^{kk'} = (2\pi)^{d} \delta(k - k') p_A(k) $$
#
# We can now non-parametrically model $p_A(k)$, with the only constraint being the positivity of the power spectrum. To enforce positivity, let us model $\gamma(k)$ where
# $$ p_A(k) \propto e^{\gamma(k)} $$
#
# $\gamma(k)$ is modeled with an Integrated Wiener Process in $l = log|k|$ coordinates. In logarithmic coordinates, the zero mode $|k| = 0$ is infinitely far away from the other modes, hence it is treated separately.
#
# The integrated Wiener process is a stochastic process that is described by the equation:
#
# $$ \frac{d^2 \gamma(k)}{dl^2} = \eta (l) $$
# where $\eta(l)$ is standard Gaussian distributed.
#
# One way to solve the differential equation is by splitting the single second order differential equation into two first order differential equations by substitution.
#
# $$ v = \frac{d \gamma(k)}{dl} $$
# $$ \frac{dv}{dl} = \eta(l) $$
#
# Integrating both equations wrt $l$ results in:
#
# \begin{equation}
#     \int_{l_0}^{l} v(l') dl' = \gamma(k) - \gamma(0)
# \end{equation}
#
# \begin{equation}
#     v(l') - v(l_0) = \int_{l_0}^{l'} \eta(l'') dl''
# \end{equation}
#
#
#
#
#
#
#

# ## Mathematical intuition for the Fluctuations parameter
#
# The two-point correlation function between two locations $x$ and $y$ is given by $S^{xy}$ which is a continuous function of the distance between the two points, $x-y$ assuming a priori statistical homogeneity.  $$ S^{xy}= C_s(x-y) \mathrel{\mathop{:}}= C_s(r)$$
#
# When this is the case, the correlation structure is diagonalized in harmonic space and is described fully by the Power Spectrum $P_s(k)$ i.e,
#
# \begin{align*}
#     S^{xy} & = (F^\dagger)^r_k P_s (k) \\
#            & = \int \frac{\mathop{dk}}{(2\pi)^u} \exp{[-ikr]} P_s(k) \\
# \end{align*}
#
# Then, the auto correlation
#
# \begin{align*}
#     S^{xx} & = \left< s^x s^{x*}\right> \\
#     & = C_s(0) = \int \frac{dk}{(2\pi)^u} e^{0} P_s(k) \\
#     \int P_s(k) \mathop{dk} & = \left< |s^x|^2 \right>
#
# \end{align*}
#
# This $P_s(k)$ is modified to a power spectrum with added fluctuations,
#
# \begin{align*}
# P_s ' = \frac{P_s}{\sqrt{\int \mathop{dk} P_s(k)}} a^2
# \end{align*}
